"""Balance Sheet Items Objects
"""

from .balance_sheet_item import BalanceSheetItem
from .balance_sheet_submain_group import (
    BalanceSheetSubMainGroupObjects as BSSubMain,
)

# Balance Sheet Item Objects
# These are objects that can be used

CASH = BalanceSheetItem(
    ID="101-0000000001",
    name="Cash",
    origin="Cash available",
    balance_sheet_sub_main_group=BSSubMain.SHORT_TERM_ASSET,
)

ACCOUNTS_RECEIVABLE = BalanceSheetItem(
    ID="101-0000000002",
    name="Accounts Receivable",
    origin="Accounts receivable (AR) is the balance"
    " of money due to a firm for goods or services"
    "delivered or used but not yet paid for by customer",
    balance_sheet_sub_main_group=BSSubMain.SHORT_TERM_ASSET,
)

ALLOWANCE_FOR_BAD_DEBTS = BalanceSheetItem(
    ID="101-0000000003",
    name="Allowance for Bad Debts",
    origin="An allowance for bad debt is a valuation account "
    "used to estimate the amount of a firm's receivables that"
    "may ultimately be uncollectible.",
    balance_sheet_sub_main_group=BSSubMain.SHORT_TERM_ASSET,
)

INVENTORY = BalanceSheetItem(
    ID="101-0000000004",
    name="Inventory",
    origin=(
        "Inventory is the raw materials used to produce"
        "goods as well as the goods that are available for sale."
        "The three types of inventory include raw materials,"
        "work-in-progress, and finished goods."
    ),
    balance_sheet_sub_main_group=BSSubMain.SHORT_TERM_ASSET,
)

SHORT_PREPAID_INSURANCE = BalanceSheetItem(
    ID="101-0000000005",
    name="Prepaid Insurance",
    origin=(
        "A prepaid expense is carried on an insurance company's"
        " balance sheet as a current asset until it is consumed."
    ),
    balance_sheet_sub_main_group=BSSubMain.SHORT_TERM_ASSET,
)

TRUCK = BalanceSheetItem(
    ID="102-0000000001",
    name="Truck",
    origin="Trucks",
    balance_sheet_sub_main_group=BSSubMain.LONG_TERM_ASSET,
)

LAND = BalanceSheetItem(
    ID="102-0000000002",
    name="Land",
    origin="Land",
    balance_sheet_sub_main_group=BSSubMain.LONG_TERM_ASSET,
)

BUILDING = BalanceSheetItem(
    ID="102-0000000003",
    name="Building",
    origin="Building",
    balance_sheet_sub_main_group=BSSubMain.LONG_TERM_ASSET,
)

PREPAID_INSURANCE = BalanceSheetItem(
    ID="101-0000000004",
    name="Prepaid Insurance",
    origin=(
        "A prepaid expense is carried on an insurance"
        "company's balance sheet"
        "as a current asset until it is consumed."
    ),
    balance_sheet_sub_main_group=BSSubMain.LONG_TERM_ASSET,
)


ACCOUNTS_PAYABLE = BalanceSheetItem(
    ID="201-0000000001",
    name="Accounts Payable",
    origin="Accounts Payable",
    balance_sheet_sub_main_group=BSSubMain.SHORT_TERM_LIABILITY,
)

NOTES_PAYABLE = BalanceSheetItem(
    ID="211-0000000001",
    name="Notes Payable",
    origin="Notes Payable",
    balance_sheet_sub_main_group=BSSubMain.LONG_TERM_LIABILITY,
)

ORIGINAL_INVESTMENT = BalanceSheetItem(
    ID="211-0000000001",
    name="Original Investment",
    origin="Original investment by owners of the company",
    balance_sheet_sub_main_group=BSSubMain.OWNERS_EQUITY,
)
RETAINED_EARNINGS = BalanceSheetItem(
    ID="211-0000000001",
    name="Retained Earnings",
    origin="Retained Earnings is the earning that"
    "is re-invested in the company",
    balance_sheet_sub_main_group=BSSubMain.OWNERS_EQUITY,
)



import csv
import json
import os
import pathlib

from sqlalchemy import Engine
from sqlalchemy import text

from xarizmi.db.models.constants import TableNamesEnum


def export_table_to_csv(
    table_name: str, output_file: str, engine: Engine
) -> None:
    """
    Exports a PostgreSQL table to a CSV file using an existing SQLAlchemy
    engine.

    Args:
        table_name (str): Name of the table to export.
        output_file (str): Path to the output CSV file.
        engine (sqlalchemy.engine.Engine): SQLAlchemy engine
        connected to the database.
    """
    with engine.connect() as connection:
        # Query the table
        query = text(f"SELECT * FROM {table_name}")
        result = connection.execute(query)

        # Fetch column names (headers)
        headers = result.keys()

        # Export to CSV
        with open(output_file, mode="w", newline="") as csv_file:
            writer = csv.writer(csv_file)

            # Write headers
            writer.writerow(headers)

            # Write rows
            for row in result:
                writer.writerow(row)

    print(f"Table {table_name} has been exported to {output_file}")


def export_table_to_json(
    table_name: str, output_file: str | pathlib.Path, engine: Engine
) -> None:
    """
    Exports a PostgreSQL table to a JSON file using SQLAlchemy with optional limit and offset.

    Args:
        table_name (str): Name of the table to export.
        output_file (str): Path to the output JSON file.
        engine (sqlalchemy.engine.Engine): SQLAlchemy engine connected to the database.
    """
    with engine.connect() as connection:
        # Build the query with optional LIMIT and OFFSET
        limit = 1000000
        for step in range(0, 10000):
            offset = step * limit

            query = f"SELECT * FROM {table_name}"
            if limit is not None:
                query += f" LIMIT {limit}"
            if offset is not None:
                query += f" OFFSET {offset}"

            # Execute the query
            result = connection.execute(text(query))

            # Fetch column names (headers)
            headers = result.keys()

            # Convert rows to a list of dictionaries
            data = [dict(zip(headers, row)) for row in result]

            if len(data) <= 0:
                break

            # Export to JSON
            with open(
                str(output_file).replace(".json", f"_{step}.json"), mode="w"
            ) as json_file:
                json.dump(data, json_file, indent=4, default=str)

        print(
            f"Table {table_name} has been exported to {output_file} (Limit: {limit}, Offset: {offset})"
        )


def export_candlestick_table_to_json(output_file: str, engine: Engine) -> None:
    with engine.connect() as connection:
        # Build the query with optional LIMIT and OFFSET
        limit = 1000000
        for step in range(0, 1000):
            offset = step * limit

            query = f"""SELECT a.*,
            b.exchange_name as exchange_name,
            b.base_currency as base_currency,
            b.quote_currency as quote_currency,
            b.fee_currency as fee_currency
            FROM {TableNamesEnum.CANDLESTICK.value} as a
            INNER JOIN {TableNamesEnum.SYMBOL.value} as b
            ON a.symbol_id = b.id
            """
            if limit is not None:
                query += f" LIMIT {limit}"
            if offset is not None:
                query += f" OFFSET {offset}"

            # Execute the query
            result = connection.execute(text(query))

            # Fetch column names (headers)
            headers = result.keys()

            # Convert rows to a list of dictionaries
            data = [dict(zip(headers, row)) for row in result]

            if len(data) <= 0:
                break

            # Export to JSON
            with open(
                str(output_file).replace(".json", f"_{step}.json"), mode="w"
            ) as json_file:
                json.dump(data, json_file, indent=4, default=str)


def export_all_xarizmi_tables_to_json(
    engine: Engine,
    parent_dir: pathlib.Path = pathlib.Path(os.getcwd()),
) -> None:
    export_table_to_json(
        TableNamesEnum.EXCHANGE.value,
        parent_dir / f"{TableNamesEnum.EXCHANGE.value}.json",
        engine=engine,
    )
    export_table_to_json(
        TableNamesEnum.SYMBOL.value,
        parent_dir / f"{TableNamesEnum.SYMBOL.value}.json",
        engine=engine,
    )
    export_candlestick_table_to_json(
        parent_dir / f"{TableNamesEnum.CANDLESTICK.value}.json",
        engine=engine,
    )
    export_table_to_json(
        TableNamesEnum.PORTFOLIO_ITEM.value,
        parent_dir / f"{TableNamesEnum.PORTFOLIO_ITEM.value}.json",
        engine=engine,
    )



import datetime

import numpy as np

from xarizmi.candlestick import Candlestick


def generate_four_random_numbers(
    lower_limit: float = 0, scale: float = 1
) -> list[float]:
    nums = lower_limit + (np.random.random(4) * scale)
    nums.sort()
    return nums.tolist()  # type: ignore


def generate_random_candlestick(
    lower_limit: float = 0,
    scale: float = 1,
    datetime: None | datetime.datetime = None,
) -> Candlestick:
    nums = generate_four_random_numbers(lower_limit=lower_limit, scale=scale)
    return Candlestick(
        close=nums[2],
        open=nums[1],
        low=nums[0],
        high=nums[3],
        volume=np.random.randint(low=100, high=200),
        amount=np.random.randint(low=100, high=200),
        datetime=datetime,
    )


def generate_random_bullish_candlestick(
    lower_limit: float = 0,
    scale: float = 1,
    datetime: datetime.datetime | None = None,
) -> Candlestick:
    nums = generate_four_random_numbers(lower_limit=lower_limit, scale=scale)
    return Candlestick(
        close=nums[2],
        open=nums[1],
        low=nums[0],
        high=nums[3],
        volume=np.random.randint(low=100, high=200),
        amount=np.random.randint(low=100, high=200),
        datetime=datetime,
    )


def generate_random_bearish_candlestick(
    lower_limit: float = 0,
    scale: float = 1,
    datetime: None | datetime.datetime = None,
) -> Candlestick:
    nums = generate_four_random_numbers(lower_limit=lower_limit, scale=scale)
    return Candlestick(
        close=nums[1],
        open=nums[2],
        low=nums[0],
        high=nums[3],
        volume=np.random.randint(low=100, high=200),
        amount=np.random.randint(low=100, high=200),
        datetime=datetime,
    )



from sqlalchemy.orm import Session

from xarizmi.models.exchange import Exchange as PyExchange

from ..models import Exchange


def upsert_exchange(exchange: PyExchange, session: Session) -> Exchange:
    """Creates exchange in db or returns it if it already exists"""
    db_exchange = (
        session.query(Exchange)
        .filter_by(
            name=exchange.name,
        )
        .first()
    )

    if not db_exchange:
        db_exchange = Exchange(
            name=exchange.name,
        )
        session.merge(db_exchange)
        session.flush()
        session.commit()
    return db_exchange


def bulk_upsert_exchanges(
    exchanges: list[PyExchange], session: Session
) -> list[Exchange]:
    db_exchanges = []
    for exchange in exchanges:
        db_exchange = upsert_exchange(exchange=exchange, session=session)
        db_exchanges.append(db_exchange)
    return db_exchanges



"""first_migrations

Revision ID: 4bcb140de81f
Revises:
Create Date: 2024-11-23 00:32:01.735301

"""

from typing import Sequence
from typing import Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "4bcb140de81f"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "xarizmi_exchange",
        sa.Column("name", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_table(
        "xarizmi_symbol",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("base_currency", sa.String(), nullable=False),
        sa.Column("quote_currency", sa.String(), nullable=False),
        sa.Column("fee_currency", sa.String(), nullable=True),
        sa.Column(
            "name",
            sa.String(),
            sa.Computed(
                "base_currency || '-' || quote_currency",
            ),
            nullable=False,
        ),
        sa.Column("exchange_name", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["exchange_name"],
            ["xarizmi_exchange.name"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "base_currency",
            "quote_currency",
            "exchange_name",
            name="uix_symbol_name_exchange",
        ),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "xarizmi_candlestick",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("symbol_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["symbol_id"],
            ["xarizmi_symbol.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("xarizmi_candlestick")
    op.drop_table("xarizmi_symbol")
    op.drop_table("xarizmi_exchange")
    # ### end Alembic commands ###



from datetime import datetime as dt

from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint

from xarizmi.db.models.base import Base

from .constants import TableNamesEnum
from .symbol import Symbol


class PortfolioItem(Base):  # type: ignore
    __tablename__ = TableNamesEnum.PORTFOLIO_ITEM.value
    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    symbol_id: Mapped[int] = mapped_column(
        Integer, ForeignKey(Symbol.id), nullable=False
    )
    symbol: Mapped[Symbol] = relationship(
        "Symbol", back_populates="portfolio_items"
    )

    market_value: Mapped[float] = mapped_column(Float, nullable=False)
    quantity: Mapped[float] = mapped_column(Float, nullable=False)
    datetime: Mapped[dt] = mapped_column(
        DateTime,
        nullable=False,
    )

    __table_args__ = (
        UniqueConstraint(
            "symbol_id",
            "datetime",
            name="uix_symbol_datetime",
        ),
    )



from sqlalchemy.orm import Session

from xarizmi.candlestick import Candlestick as pyCandlestick
from xarizmi.db.models.candlestick import CandleStick


def upsert_candlestick(
    candlestick: pyCandlestick, symbol_id: int, session: Session
) -> CandleStick:
    """Creates candlestick in db or returns it if it already exists"""
    db_candlestick = (
        session.query(CandleStick)
        .filter_by(
            symbol_id=symbol_id,
            interval_type=candlestick.interval_type,
            interval=candlestick.interval,
        )
        .first()
    )

    if db_candlestick:
        db_candlestick.open = candlestick.open
        db_candlestick.close = candlestick.close
        db_candlestick.low = candlestick.low
        db_candlestick.high = candlestick.high
        db_candlestick.volume = candlestick.volume
        db_candlestick.amount = candlestick.amount  # type: ignore
        db_candlestick.datetime = candlestick.datetime  # type: ignore
    else:
        db_candlestick = CandleStick(
            symbol_id=symbol_id,
            open=candlestick.open,
            close=candlestick.close,
            high=candlestick.high,
            low=candlestick.low,
            datetime=candlestick.datetime,
            interval=candlestick.interval,
            interval_type=candlestick.interval_type,
            volume=candlestick.volume,
            amount=candlestick.amount,
        )
        session.merge(db_candlestick)
    session.commit()
    return db_candlestick



"""add_uniqueness_constraints

Revision ID: e726c627ded5
Revises: 0f106fabd22f
Create Date: 2024-11-23 10:55:20.681791

"""

from typing import Sequence
from typing import Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "e726c627ded5"
down_revision: Union[str, None] = "0f106fabd22f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "xarizmi_candlestick",
        "interval_type",
        existing_type=postgresql.INTERVAL(),
        type_=sa.Enum(
            "MIN_1",
            "MIN_3",
            "MIN_5",
            "MIN_15",
            "MIN_30",
            "HOUR_1",
            "HOUR_2",
            "HOUR_4",
            "HOUR_6",
            "HOUR_8",
            "HOUR_12",
            "DAY_1",
            "DAY_7",
            "DAY_14",
            "WEEK_1",
            "MONTH_1",
            name="interval_type_enum",
        ),
        existing_nullable=False,
    )
    op.create_unique_constraint(
        "uix_symbol_interval",
        "xarizmi_candlestick",
        ["symbol_id", "interval_type", "interval"],
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "uix_symbol_interval", "xarizmi_candlestick", type_="unique"
    )
    op.alter_column(
        "xarizmi_candlestick",
        "interval_type",
        existing_type=sa.Enum(
            "MIN_1",
            "MIN_3",
            "MIN_5",
            "MIN_15",
            "MIN_30",
            "HOUR_1",
            "HOUR_2",
            "HOUR_4",
            "HOUR_6",
            "HOUR_8",
            "HOUR_12",
            "DAY_1",
            "DAY_7",
            "DAY_14",
            "WEEK_1",
            "MONTH_1",
            name="interval_type_enum",
        ),
        type_=postgresql.INTERVAL(),
        existing_nullable=False,
    )
    # ### end Alembic commands ###



from sqlalchemy import select
from sqlalchemy.orm import Session
from sqlalchemy.sql import and_

from xarizmi.db.actions.symbol import get_symbol
from xarizmi.db.models.order import Order
from xarizmi.enums import OrderStatusEnum
from xarizmi.enums import SideEnum
from xarizmi.models import Symbol as PySymbol
from xarizmi.models.orders import Order as PyOrder


def get_unique_order_in_db(
    session: Session,
    symbol: PySymbol,
    order_id: str,
) -> Order | None:
    query = select(
        Order,  # All columns from Order
    )

    # Add filters
    filters = []
    db_symbol = get_symbol(symbol=symbol, session=session)
    filters.append(Order.symbol_id == db_symbol.id)
    filters.append(Order.order_id == order_id)

    if filters:
        query = query.where(and_(*filters))

    result = session.execute(query).one_or_none()
    if result:
        return result._asdict()["Order"]  # type: ignore
    else:
        return None


def get_unique_order(
    session: Session,
    symbol: PySymbol,
    order_id: str,
) -> PyOrder | None:
    db_order = get_unique_order_in_db(
        session=session, symbol=symbol, order_id=order_id
    )
    if db_order:
        return PyOrder(
            symbol=symbol,
            order_id=db_order.order_id,
            price=db_order.price,
            amount=db_order.amount,
            side=db_order.side,
            status=db_order.status,
        )
    else:
        return None


def get_orders(
    session: Session,
    symbol: PySymbol,
    filter_by_status: list[OrderStatusEnum] | None = None,
    filter_by_side: SideEnum | None = None,
) -> list[PyOrder]:
    query = select(
        Order,  # All columns from Order
    )

    # Add filters
    filters = []
    db_symbol = get_symbol(symbol=symbol, session=session)
    filters.append(Order.symbol_id == db_symbol.id)
    if filter_by_side:
        filters.append(Order.side == filter_by_side)
    if filter_by_status:
        filters.append(Order.status.in_(filter_by_status))

    if filters:
        query = query.where(and_(*filters))

    db_orders: list[Order] = [
        item._asdict()["Order"] for item in session.execute(query).all()
    ]

    return [
        PyOrder(
            symbol=symbol,
            order_id=db_order.order_id,
            price=db_order.price,
            amount=db_order.amount,
            side=db_order.side,
            status=db_order.status,
        )
        for db_order in db_orders
    ]


def get_active_orders(
    session: Session,
    symbol: PySymbol,
    filter_by_side: SideEnum | None = None,
) -> list[PyOrder]:
    return get_orders(
        session=session,
        symbol=symbol,
        filter_by_status=[
            OrderStatusEnum.ACTIVE,
            OrderStatusEnum.PARTIALLY_FILLED,
        ],
        filter_by_side=filter_by_side,
    )



from .base import Base
from .candlestick import CandleStick
from .exchange import Exchange
from .order import Order
from .portfolio import PortfolioItem
from .symbol import Symbol

__all__ = [
    "Base",
    "Symbol",
    "Exchange",
    "CandleStick",
    "PortfolioItem",
    "Order",
]



from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Computed
from sqlalchemy.schema import UniqueConstraint

from xarizmi.db.models.base import Base

from .constants import TableNamesEnum
from .exchange import Exchange


class Symbol(Base):  # type: ignore
    __tablename__ = TableNamesEnum.SYMBOL.value
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    base_currency: Mapped[str] = mapped_column(String, nullable=False)
    quote_currency: Mapped[str] = mapped_column(String, nullable=False)
    fee_currency: Mapped[str] = mapped_column(String, nullable=True)
    name: Mapped[str] = mapped_column(
        String, Computed("base_currency || '-' || quote_currency"), unique=True
    )
    exchange_name: Mapped[str] = mapped_column(
        String, ForeignKey(Exchange.name), nullable=False
    )

    # Establish a many-to-one relationship
    exchange: Mapped[Exchange] = relationship(
        "Exchange", back_populates="symbols"
    )

    candlesticks: Mapped[list["CandleStick"]] = relationship(  # type: ignore  # noqa: F821,E501
        "CandleStick", back_populates="symbol"
    )

    portfolio_items: Mapped[list["PortfolioItem"]] = relationship(  # type: ignore  # noqa: F821,E501
        "PortfolioItem", back_populates="symbol"
    )

    orders: Mapped[list["Order"]] = relationship(  # type: ignore  # noqa: F821,E501
        "Order", back_populates="symbol"
    )

    @property
    def symbol(self) -> str:
        return f"{self.base_currency}-{self.quote_currency}"

    __table_args__ = (
        UniqueConstraint(
            "base_currency",
            "quote_currency",
            "exchange_name",
            name="uix_symbol_name_exchange",
        ),
    )



import json
import pathlib
from pathlib import Path
from typing import Callable

from sqlalchemy.orm import Session

from xarizmi.candlestick import Candlestick
from xarizmi.candlestick import CandlestickList
from xarizmi.db.actions.candlestick import upsert_candlestick
from xarizmi.db.actions.exchange import bulk_upsert_exchanges
from xarizmi.db.actions.symbol import bulk_upsert_symbols
from xarizmi.db.actions.symbol import get_symbol
from xarizmi.models.exchange import Exchange
from xarizmi.models.exchange import ExchangeList
from xarizmi.models.symbol import Symbol
from xarizmi.models.symbol import SymbolList


def list_files_with_prefix(
    parent_dir: str | pathlib.Path, file_prefix: str
) -> list[Path]:
    if isinstance(parent_dir, str):
        parent_dir = Path(parent_dir)

    if not parent_dir.is_dir():
        raise ValueError(f"The directory {parent_dir} does not exist.")

    # List all files in the directory that start with the given prefix
    files = [
        file
        for file in parent_dir.iterdir()
        if file.is_file() and file.name.startswith(file_prefix)
    ]

    return files


def load_pydantic_model_from_db_model_exported_json_file(
    parent_dir: pathlib.Path,
    converter_function: Callable,
    file_prefix: str = "xarizmi_exchange_",
    transformer: Callable | None = None,
    insert_in_db: bool = False,
) -> list[Exchange]:
    files = list_files_with_prefix(
        parent_dir=parent_dir, file_prefix=file_prefix
    )
    for filepath in files:
        with open(filepath) as f:
            data = json.load(f)

        if transformer is not None:
            data = map(transformer, data)
        pydantic_items_wrapper = converter_function(items=data)
        yield pydantic_items_wrapper.items


def load_pydantic_exchange_model_from_db_model_exported_json_file(
    parent_dir,
    file_prefix: str = "xarizmi_exchange_",
    insert_in_db: bool = False,
    session: None | Session = None,
) -> list[Exchange]:
    generator = load_pydantic_model_from_db_model_exported_json_file(
        parent_dir, ExchangeList, file_prefix
    )
    exchanges = []
    for items in generator:
        exchanges += items
    if insert_in_db is True:
        if session is None:
            raise RuntimeError("Session is not provided!")
        bulk_upsert_exchanges(exchanges, session=session)
    return exchanges


def load_pydantic_symbol_model_from_db_model_exported_json_file(
    parent_dir,
    file_prefix: str = "xarizmi_symbol_",
    insert_in_db: bool = False,
    session: None | Session = None,
) -> list[Symbol]:

    def transformer(symbol_item: dict[str, str]) -> dict[str, str]:
        symbol_item["exchange"] = symbol_item["exchange_name"]
        return symbol_item

    generator = load_pydantic_model_from_db_model_exported_json_file(
        parent_dir,
        SymbolList.build,
        file_prefix,
        transformer=transformer,
    )
    symbols = []
    for items in generator:
        symbols += items
    if insert_in_db is True:
        bulk_upsert_symbols(symbols, session=session)
    return symbols


def load_pydantic_candlestick_model_from_db_model_exported_json_file(
    parent_dir: str,
    file_prefix: str = "xarizmi_candlestick_",
    insert_in_db: bool = False,
    session: None | Session = None,
    keep_in_memory: bool = False,
) -> list[Candlestick]:

    def transformer(item: dict[str, str]) -> dict[str, str]:

        item["symbol"] = Symbol.build(
            base_currency=item["base_currency"],
            quote_currency=item["quote_currency"],
            fee_currency=item["fee_currency"],
            exchange=item["exchange_name"],
        )

        item["exchange"] = Exchange(name=item["exchange_name"])

        # interval type
        if item["interval_type"] == "MIN_1":
            item["interval_type"] = "1min"
        if item["interval_type"] == "MIN_3":
            item["interval_type"] = "3min"
        if item["interval_type"] == "MIN_5":
            item["interval_type"] = "5min"
        if item["interval_type"] == "MIN_15":
            item["interval_type"] = "15min"
        if item["interval_type"] == "MIN_30":
            item["interval_type"] = "30min"
        if item["interval_type"] == "HOUR_1":
            item["interval_type"] = "1hour"
        if item["interval_type"] == "HOUR_2":
            item["interval_type"] = "2hour"
        if item["interval_type"] == "HOUR_3":
            item["interval_type"] = "3hour"
        if item["interval_type"] == "HOUR_4":
            item["interval_type"] = "4hour"
        if item["interval_type"] == "HOUR_6":
            item["interval_type"] = "6hour"
        if item["interval_type"] == "HOUR_8":
            item["interval_type"] = "8hour"
        if item["interval_type"] == "HOUR_12":
            item["interval_type"] = "12hour"
        if item["interval_type"] == "DAY_1":
            item["interval_type"] = "1day"
        if item["interval_type"] == "DAY_7":
            item["interval_type"] = "7day"
        if item["interval_type"] == "DAY_14":
            item["interval_type"] = "14day"
        if item["interval_type"] == "WEEK_1":
            item["interval_type"] = "7day"
        if item["interval_type"] == "WEEK_2":
            item["interval_type"] = "14day"
        if item["interval_type"] == "MONTH_1":
            item["interval_type"] = "1month"
        return item

    candlesticks = []
    generator = load_pydantic_model_from_db_model_exported_json_file(
        parent_dir,
        CandlestickList,
        file_prefix,
        transformer=transformer,
    )
    cached_symbol_ids: dict[Symbol, int] = {}

    for items in generator:
        if keep_in_memory is True:
            candlesticks += items
        if insert_in_db is True:
            for candlestick in items:
                symbol_id = cached_symbol_ids.get(candlestick.symbol, None)
                if symbol_id is None:
                    symbol_id = get_symbol(
                        symbol=candlestick.symbol, session=session
                    ).id
                    cached_symbol_ids[candlestick.symbol] = symbol_id
                upsert_candlestick(
                    candlestick=candlestick,
                    symbol_id=symbol_id,
                    session=session,
                )
        print("Batch of candlestick items inserted into db ...")

    return candlesticks



"""add_portfolio_model

Revision ID: c63cb9ba06b2
Revises: dfe454fc4858
Create Date: 2024-11-25 23:20:40.706753

"""

from typing import Sequence
from typing import Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "c63cb9ba06b2"
down_revision: Union[str, None] = "dfe454fc4858"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "xarizmi_portfolio_item",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("symbol_id", sa.Integer(), nullable=False),
        sa.Column("market_value", sa.Float(), nullable=False),
        sa.Column("quantity", sa.Float(), nullable=False),
        sa.Column("datetime", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["symbol_id"],
            ["xarizmi_symbol.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "symbol_id", "datetime", name="uix_symbol_datetime"
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("xarizmi_portfolio_item")
    # ### end Alembic commands ###



from datetime import datetime

from sqlalchemy import desc
from sqlalchemy.orm import Session

from xarizmi.db.models.portfolio import PortfolioItem
from xarizmi.db.models.symbol import Symbol
from xarizmi.models.portfolio import Portfolio as PyPortfolio
from xarizmi.models.portfolio import PortfolioItem as PyPortfolioItem


def get_portfolio_items_between_dates(
    session: Session, start_date: datetime, end_date: datetime
) -> PyPortfolio:
    results = (
        session.query(
            PortfolioItem.market_value,
            PortfolioItem.quantity,
            PortfolioItem.datetime,
            Symbol.base_currency,
            Symbol.quote_currency,
            Symbol.fee_currency,
            Symbol.exchange_name,
        )
        .join(Symbol, PortfolioItem.symbol_id == Symbol.id)
        .filter(PortfolioItem.datetime.between(start_date, end_date))
        .order_by(desc(PortfolioItem.market_value))
    ).all()
    results = [item._asdict() for item in results]  # type: ignore

    return PyPortfolio(
        items=[
            PyPortfolioItem.model_validate(
                {
                    "symbol": {
                        "base_currency": {
                            "name": item["base_currency"],  # type: ignore
                        },
                        "quote_currency": {
                            "name": item["quote_currency"],  # type: ignore
                        },
                        "fee_currency": {
                            "name": item["fee_currency"],  # type: ignore
                        },
                        "exchange": {
                            "name": item["exchange_name"],  # type: ignore
                        },
                    },
                    "market_value": item["market_value"],  # type: ignore
                    "quantity": item["quantity"],  # type: ignore
                    "datetime": item["datetime"],  # type: ignore
                }
            )
            for item in results
        ]
    )



from xarizmi.candlestick import CandlestickChart


def example_simple_candlestick_plot_no_volume() -> None:

    candles_data = [
        {
            "datetime": "2024-12-18",
            "open": 100,
            "close": 103,
            "high": 105,
            "low": 98,
            "volume": 1,
        },
        {
            "datetime": "2024-12-19",
            "open": 102,
            "close": 106,
            "high": 107,
            "low": 101,
            "volume": 1,
        },
        {
            "datetime": "2024-12-20",
            "open": 107,
            "close": 103,
            "high": 108,
            "low": 101,
            "volume": 1,
        },
    ]

    candlesticks = CandlestickChart.model_validate({"candles": candles_data})

    candlesticks.save_simple_plot_no_volume("candlestick_chart.svg")


if __name__ == "__main__":
    example_simple_candlestick_plot_no_volume()



from sqlalchemy.orm import Session

from xarizmi.models.symbol import Symbol as PySymbol

from ..models import Symbol


def get_symbol(symbol: PySymbol, session: Session) -> Symbol:
    db_symbol = (
        session.query(Symbol)
        .filter_by(
            base_currency=symbol.base_currency.name,
            quote_currency=symbol.quote_currency.name,
            exchange_name=symbol.exchange.name,  # type: ignore
        )
        .first()
    )
    if db_symbol is None:
        raise RuntimeError("couldn't find the symbol in DB table!")
    return db_symbol


def upsert_symbol(symbol: PySymbol, session: Session) -> Symbol:
    """Creates symbol in db or returns it if it already exists"""
    db_symbol = (
        session.query(Symbol)
        .filter_by(
            base_currency=symbol.base_currency.name,
            quote_currency=symbol.quote_currency.name,
            exchange_name=symbol.exchange.name,  # type: ignore
        )
        .first()
    )

    if not db_symbol:
        db_symbol = Symbol(
            base_currency=symbol.base_currency.name,
            quote_currency=symbol.quote_currency.name,
            fee_currency=symbol.fee_currency.name,
            exchange_name=symbol.exchange.name,  # type: ignore
        )
        session.merge(db_symbol)
        session.flush()
    return db_symbol


def bulk_upsert_symbols(
    symbols: list[PySymbol], session: Session
) -> list[Symbol]:
    db_symbols = []
    for symbol in symbols:
        db_symbol = upsert_symbol(symbol, session=session)
        db_symbols += [db_symbol]
    return db_symbols



import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import fsolve  # type: ignore

from xarizmi.fundamentals.data_models import FormulaTuple


def generate_discounted_cash_formula(
    year: int, cf: float, kind: str | None = None
) -> FormulaTuple:
    if kind is None:
        formula_string = f"{cf}/(1+r)^{year}"

        def discount_cash_formula(r: float) -> float:
            return cf / (1 + r) ** year

        return FormulaTuple(discount_cash_formula, formula_string)
    elif kind == "perpetuity":
        formula_string = f"{cf}/r/(1+r)^{year - 1}"

        def discount_cash_formula(r: float) -> float:
            return cf / r / (1 + r) ** (year - 1)

        return FormulaTuple(discount_cash_formula, formula_string)
    else:
        raise NotImplementedError()


def generate_terminal_compounding_cash_formula(
    year: int, cf: float, terminal_year: int, kind: str | None = None
) -> FormulaTuple:
    if kind is None:
        formula_string = f"{cf} * (1+r)^{terminal_year - year}"

        def compounding_formula(r: float) -> float:
            return cf * (1 + r) ** (terminal_year - year)

        return FormulaTuple(compounding_formula, formula_string)
    elif kind == "perpetuity":
        formula_string = f"{cf}/r * (1+r)^{terminal_year - year - 1}"

        def compounding_formula(r: float) -> float:
            return cf / r * (1 + r) ** (terminal_year - year - 1)

        return FormulaTuple(compounding_formula, formula_string)
    else:
        raise NotImplementedError()


class IRR:

    def __init__(
        self,
        years: list[int],
        cfs: list[float],
        kinds: list[str | None] | None = None,
        opportunity_cost: None | float = None,
    ) -> None:
        if kinds is None:
            kinds = [None] * len(years)
        assert len(years) == len(cfs) and len(cfs) == len(kinds)
        self._years = years
        self._cfs = cfs
        self._kinds = kinds
        self._opportunity_cost = opportunity_cost
        self._formula_tuple = self._process()

    @property
    def years(self) -> list[int]:
        return self._years

    @property
    def opportunity_cost(self) -> None | float:
        return self._opportunity_cost

    @property
    def cfs(self) -> list[float]:
        return self._cfs

    @property
    def kinds(self) -> list[str | None]:
        return self._kinds

    @property
    def formula(self) -> FormulaTuple:
        return self._formula  # type: ignore

    @property
    def formula_string(self) -> str:
        return self._formula_string  # type: ignore

    def _process(self) -> FormulaTuple:
        func_list = []
        func_string_list = []
        for year, cf, kind in zip(self.years, self.cfs, self.kinds):
            func_list.append(
                generate_discounted_cash_formula(year, cf, kind).formula
            )
            func_string_list.append(
                generate_discounted_cash_formula(year, cf, kind).formula_string
            )
        formula_string = "+".join(func_string_list)

        def formula(r: float) -> float:
            return sum(f(r) for f in func_list)  # type: ignore

        formula_tuple = FormulaTuple(formula, formula_string)
        self._formula = formula_tuple.formula
        self._formula_string = formula_tuple.formula_string
        return formula_tuple

    def find_mirr(self, precision: int = 4) -> float:
        """Returns Modified IRR"""
        if self.opportunity_cost is None:
            raise ValueError(
                "To calculate MIRR you need to provide"
                " opportunity cost in IRR constructor"
            )
        func_list = []
        terminal_year = max(self.years)
        initial_investment = abs(
            sum(cf for year, cf in zip(self.years, self.cfs) if year == 0)
        )
        for year, cf, kind in zip(self.years, self.cfs, self.kinds):
            if year == 0:
                continue
            func_list.append(
                generate_terminal_compounding_cash_formula(
                    year, cf, terminal_year, kind
                ).formula
            )

        def F(r: float) -> float:
            return sum(f(r) for f in func_list)  # type: ignore

        mirr = (F(self.opportunity_cost) / initial_investment) ** (
            1 / terminal_year
        ) - 1
        return round(mirr, precision)  # type: ignore

    def __str__(self) -> str:
        return f"Formula = {self.formula_string}"

    def __repr__(self) -> str:
        s = "IRRFormula(\n"
        s += f"years={self.years},\n"
        s += f"cfs={self.cfs},\n"
        s += f"kinds={self.kinds},\n"
        s += ")"
        return s

    def find(self, initial_guess: float = 0.1, precision: int = 4) -> float:
        solution = fsolve(self._formula, initial_guess)[0]
        return round(solution, precision)  # type: ignore

    def find_all(
        self,
        min_r: float | int = 0,
        max_r: float | int = 1,
        max_n_roots: int = 2,
        precision: int = 4,
    ) -> list[set[float]]:
        initial_guesses = np.linspace(min_r, max_r, max_n_roots)
        solutions = fsolve(self._formula, initial_guesses)
        solutions = [round(item, precision) for item in solutions]
        return list(set(solutions))

    def get_yield_curve(
        self, min_r: float | int = 0, max_r: float | int = 1, points: int = 100
    ) -> tuple[list[float], list[float]]:
        assert min_r < max_r
        assert type(points) is int
        xs = np.linspace(min_r, max_r, points)
        ys = self.formula(xs)  # type: ignore
        return xs.tolist(), ys.tolist()

    def plot(
        self,
        min_r: float = 0,
        max_r: float = 1,
        points: int = 100,
        figsize: tuple[int | float, int | float] = (10, 5),
        title: str = "Yield Curve",
        plot_label: str = "Yield",
        y_label: str = "NPV",
        x_label: str = "Return Rate",
        color: str = "blue",
        x_in_percentage: bool = True,
        grid: bool = True,
    ) -> None:
        xs, ys = self.get_yield_curve(min_r=min_r, max_r=max_r, points=points)
        if x_in_percentage:
            xs = [100 * x for x in xs]
            x_label = x_label + " (%)"
        fig, ax = plt.subplots(figsize=figsize)
        _ = ax.plot(xs, ys, label=plot_label, color=color)
        _ = ax.set_title(title)
        _ = ax.set_ylabel(y_label)
        _ = ax.set_xlabel(x_label)
        if grid is True:
            ax.grid()
        return fig, ax  # type: ignore



from sqlalchemy.orm import Session

from xarizmi.db.models.order import Order
from xarizmi.enums import OrderStatusEnum
from xarizmi.models import Symbol as PySymbol

from .read import get_unique_order_in_db


def delete_unique_order(
    session: Session,
    symbol: PySymbol,
    order_id: str,
) -> None:
    db_order = get_unique_order_in_db(
        session=session, symbol=symbol, order_id=order_id
    )
    session.delete(db_order)


def delete_all_cancelled_orders(
    session: Session,
) -> None:
    session.query(Order).where(
        Order.status == OrderStatusEnum.CANCELLED
    ).delete()



def calculate_return_rate_from_price(
    old_price: float, new_price: float
) -> float:
    return new_price / old_price - 1


def calculate_compound_return_rate(
    *, n_periods: int, return_rate_per_period: float
) -> float:
    return (1 + return_rate_per_period) ** n_periods - 1


def calculate_average_compound_return_rate_per_period(
    *, n_periods: int, total_return_rate: float
) -> float:
    return (1 + total_return_rate) ** (1 / n_periods) - 1  # type: ignore



from sqlalchemy.orm import Session

from xarizmi.db.actions.symbol import get_symbol
from xarizmi.db.models.portfolio import PortfolioItem
from xarizmi.models.portfolio import Portfolio as PyPortfolio
from xarizmi.models.portfolio import PortfolioItem as PyPortfolioItem


def upsert_portfolio_item(
    portfolio_item: PyPortfolioItem, session: Session
) -> PortfolioItem:
    """Creates portfolio_item in db or update and returns it if it already
    exists"""
    symbol = get_symbol(symbol=portfolio_item.symbol, session=session)
    db_portfolio_item = (
        session.query(PortfolioItem)
        .filter_by(
            datetime=portfolio_item.datetime,
            symbol_id=symbol.id,
        )
        .first()
    )

    if not db_portfolio_item:
        db_portfolio_item = PortfolioItem(
            symbol_id=symbol.id,
            market_value=portfolio_item.market_value,
            quantity=portfolio_item.quantity,
            datetime=portfolio_item.datetime,
        )
        session.merge(db_portfolio_item)
        session.flush()
    else:
        db_portfolio_item.quantity = portfolio_item.market_value
        db_portfolio_item.market_value = portfolio_item.market_value

    session.commit()
    return db_portfolio_item


def bulk_upsert_portfolio_item(
    portfolio_items: list[PyPortfolioItem], session: Session
) -> list[PortfolioItem]:
    db_portfolio_items = []
    for portfolio_item in portfolio_items:
        db_portfolio_item = upsert_portfolio_item(
            portfolio_item=portfolio_item, session=session
        )
        db_portfolio_items.append(db_portfolio_item)
    return db_portfolio_items


def upsert_portfolio(
    portfolio: PyPortfolio, session: Session
) -> list[PortfolioItem]:
    return bulk_upsert_portfolio_item(
        portfolio_items=portfolio.items, session=session
    )



# Download sample data for several stocks traded in Toronto Stock Exchange data

import os
import pathlib

from xarizmi.opendata.yahoo_finance import YahooFinanceDailyDataClient

SYMBOLS_LIST: list[str] = [
    "T.TO",
    "SHOP.TO",
    "CNR.TO",
    "CNQ.TO",
    "ENB.TO",
    "AC.TO",
    "OTEX.TO",
    "TRP.TO",
    "TFII.TO",
    "LSPD.TO",
]


def setup_data_directory() -> pathlib.Path:

    data_directory = (
        pathlib.Path(".")
        / "data"
        # / datetime.date.today().strftime("%Y-%m-%d")
    )
    if not data_directory.exists():
        data_directory.mkdir(parents=True)

    return data_directory


def download_daily_symbols(parent_directory: pathlib.Path) -> None:
    for symbol in SYMBOLS_LIST:
        YahooFinanceDailyDataClient.download_full_data(
            symbol=symbol,
            filepath=os.path.join(parent_directory, symbol + ".json"),
        )


def main() -> None:
    data_directory = setup_data_directory()
    download_daily_symbols(data_directory)



from xarizmi.fundamentals.time_value_money import time_value_money

value = time_value_money(
    impatience_to_consume=0.05,
    inflation=0.1,
    risk=0.045,
)

print(f"Time value of money is {value}")
print(f"A $1000 one year later worth ${int(value * 1000 + 1000)} now")



from sqlalchemy.orm import Session

from xarizmi.db.actions.symbol import get_symbol
from xarizmi.db.models.order import Order
from xarizmi.models.orders import Order as PyOrder


def upsert_order(
    order: PyOrder, session: Session, symbol_id: int | None = None
) -> Order:
    """Creates order in db or returns it if it already exists"""
    if symbol_id is None:
        symbol_id = get_symbol(order.symbol, session=session).id
    db_order = (
        session.query(Order)
        .filter_by(
            symbol_id=symbol_id,
            order_id=order.order_id,
        )
        .first()
    )

    if db_order:
        db_order.amount = order.amount
        db_order.price = order.price
        db_order.status = order.status
        db_order.side = order.side
    else:
        db_order = Order(
            symbol_id=symbol_id,
            order_id=order.order_id,
            price=order.price,
            amount=order.amount,
            status=order.status,
            side=order.side,
        )
        session.merge(db_order)
    session.commit()
    return db_order



import datetime

from pydantic import BaseModel
from pydantic import field_validator

from xarizmi.utils.math import divide_with_no_zero_division_error

from .symbol import Symbol


class PortfolioItem(BaseModel):
    symbol: Symbol
    market_value: float
    quantity: float
    datetime: datetime.datetime

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PortfolioItem):
            raise NotImplementedError(
                f"Can't compare objects of type {type(other)} and {type(self)}"
            )
        return (
            self.symbol == other.symbol
            and self.market_value == other.market_value
            and self.quantity == other.quantity
            and self.datetime == other.datetime
        )


class PortfolioItemDifference(BaseModel):
    symbol: Symbol
    delta_market_value: float
    delta_quantity: float
    delta_datetime: datetime.timedelta


class PortfolioItemRatio(BaseModel):
    symbol: Symbol
    market_value_ratio: float
    quantity_ratio: float
    datetime_ratio_in_days: float

    @property
    def return_ratio(self) -> float:
        return self.market_value_ratio / self.quantity_ratio


class PortfolioRatio(BaseModel):
    items: list[PortfolioItemRatio]

    @property
    def portfolio_datetime_ratio_in_days(self) -> float:
        return self.items[0].datetime_ratio_in_days

    def __len__(self) -> int:
        return len(self.items)

    @field_validator("items", mode="before")
    @classmethod
    def datetime_must_be_same(
        cls, items: list[PortfolioItemRatio]
    ) -> list[PortfolioItemRatio]:
        if len(items) == 0:
            return items
        portfolio_datetime_ratio_in_days = items[0].datetime_ratio_in_days
        for item in items:
            if item.datetime_ratio_in_days != portfolio_datetime_ratio_in_days:
                raise ValueError(
                    "PortfolioItemRatio items should have same datetime_ratio"
                )
        return items

    def __getitem__(self, symbol: Symbol) -> PortfolioItemRatio:
        items = list(filter(lambda x: x.symbol == symbol, self.items))
        if items:
            return items[0]
        else:
            return PortfolioItemRatio(
                symbol=symbol,
                market_value_ratio=1,
                quantity_ratio=1,
                datetime_ratio_in_days=self.portfolio_datetime_ratio_in_days,
            )


class PortfolioDifference(BaseModel):
    items: list[PortfolioItemDifference]

    @property
    def portfolio_difference_delta_datetime(self) -> datetime.timedelta:
        return self.items[0].delta_datetime

    @field_validator("items", mode="before")
    @classmethod
    def datetime_must_be_same(
        cls, items: list[PortfolioItemDifference]
    ) -> list[PortfolioItemDifference]:
        if len(items) == 0:
            return items
        portfolio_difference_datetime = items[0].delta_datetime
        for item in items:
            if item.delta_datetime != portfolio_difference_datetime:
                raise ValueError(
                    "PortfolioDifference items should have same delta_datetime"
                )
        return items

    def __getitem__(self, symbol: Symbol) -> PortfolioItemDifference:
        items = list(filter(lambda x: x.symbol == symbol, self.items))
        if items:
            return items[0]
        else:
            return PortfolioItemDifference(
                symbol=symbol,
                delta_market_value=0,
                delta_quantity=0,
                delta_datetime=self.portfolio_difference_delta_datetime,
            )


class Portfolio(BaseModel):
    items: list[PortfolioItem]

    @field_validator("items", mode="before")
    @classmethod
    def datetime_must_be_same(
        cls, items: list[PortfolioItem]
    ) -> list[PortfolioItem]:
        if len(items) == 0:
            raise ValueError("Portfolio cannot be empty!")
        portfolio_datetime = items[0].datetime
        for item in items:
            if item.datetime != portfolio_datetime:
                raise ValueError("Portfolio items should be in same datetime")
        return items

    @property
    def portfolio_datetime(self) -> datetime.datetime:
        return self.items[0].datetime

    def __getitem__(self, symbol: Symbol) -> PortfolioItem:
        items = list(filter(lambda x: x.symbol == symbol, self.items))
        if items:
            return items[0]
        else:
            return PortfolioItem(
                symbol=symbol,
                market_value=0,
                quantity=0,
                datetime=self.portfolio_datetime,
            )

    def __sub__(self, other: "Portfolio") -> PortfolioDifference:
        """Returns the difference between two portfolio"""
        items: list[PortfolioItemDifference] = []

        symbols = set(
            [self_item.symbol for self_item in self.items]
            + [other_item.symbol for other_item in other.items]
        )
        for symbol in symbols:
            self_item = self[symbol]
            other_item = other[symbol]
            items.append(
                PortfolioItemDifference(
                    symbol=symbol,
                    delta_market_value=self_item.market_value
                    - other_item.market_value,
                    delta_quantity=self_item.quantity - other_item.quantity,
                    delta_datetime=self_item.datetime - other_item.datetime,
                )
            )
        return PortfolioDifference(items=items)

    def __truediv__(self, other: "Portfolio") -> PortfolioRatio:
        """Returns the ratio of two portfolio"""
        items: list[PortfolioItemRatio] = []

        symbols = set(
            [self_item.symbol for self_item in self.items]
            + [other_item.symbol for other_item in other.items]
        )
        for symbol in symbols:
            self_item = self[symbol]
            other_item = other[symbol]
            items.append(
                PortfolioItemRatio(
                    symbol=symbol,
                    market_value_ratio=divide_with_no_zero_division_error(
                        self_item.market_value, other_item.market_value
                    ),
                    quantity_ratio=divide_with_no_zero_division_error(
                        self_item.quantity, other_item.quantity
                    ),
                    datetime_ratio_in_days=(
                        self_item.datetime - other_item.datetime
                    )
                    / datetime.timedelta(days=1),
                )
            )
        return PortfolioRatio(items=items)

    def __add__(self, other: PortfolioDifference) -> "Portfolio":
        """Returns a new portfolio"""
        items: list[PortfolioItem] = []

        symbols = set(
            [self_item.symbol for self_item in self.items]
            + [other_item.symbol for other_item in other.items]
        )
        for symbol in symbols:
            self_item = self[symbol]
            other_item = other[symbol]
            items.append(
                PortfolioItem(
                    symbol=symbol,
                    market_value=self_item.market_value
                    + other_item.delta_market_value,
                    quantity=self_item.quantity + other_item.delta_quantity,
                    datetime=self_item.datetime + other_item.delta_datetime,
                )
            )
        return Portfolio(items=items)



"""add_uniqueness_constraints_on_exchange

Revision ID: 258ee56e7873
Revises: e726c627ded5
Create Date: 2024-11-23 11:38:06.827226

"""

from typing import Sequence
from typing import Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "258ee56e7873"
down_revision: Union[str, None] = "e726c627ded5"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "xarizmi_candlestick",
        "interval_type",
        existing_type=postgresql.INTERVAL(),
        type_=sa.Enum(
            "MIN_1",
            "MIN_3",
            "MIN_5",
            "MIN_15",
            "MIN_30",
            "HOUR_1",
            "HOUR_2",
            "HOUR_4",
            "HOUR_6",
            "HOUR_8",
            "HOUR_12",
            "DAY_1",
            "DAY_7",
            "DAY_14",
            "WEEK_1",
            "MONTH_1",
            name="interval_type_enum",
        ),
        existing_nullable=False,
    )
    op.create_unique_constraint(None, "xarizmi_exchange", ["name"])
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "xarizmi_exchange", type_="unique")  # noqa: E501
    op.alter_column(
        "xarizmi_candlestick",
        "interval_type",
        existing_type=sa.Enum(
            "MIN_1",
            "MIN_3",
            "MIN_5",
            "MIN_15",
            "MIN_30",
            "HOUR_1",
            "HOUR_2",
            "HOUR_4",
            "HOUR_6",
            "HOUR_8",
            "HOUR_12",
            "DAY_1",
            "DAY_7",
            "DAY_14",
            "WEEK_1",
            "MONTH_1",
            name="interval_type_enum",
        ),
        type_=postgresql.INTERVAL(),
        existing_nullable=False,
    )
    # ### end Alembic commands ###



"""Module for each item in balance sheet
"""

from pydantic import BaseModel

from .balance_sheet_main_group import BalanceSheetMainGroup
from .balance_sheet_main_group import BalanceSheetMainGroupObjects as BSMain
from .balance_sheet_submain_group import BalanceSheetSubMainGroup
from .balance_sheet_submain_group import (
    BalanceSheetSubMainGroupObjects as BSSubMain,
)


class BalanceSheetItem(BaseModel):
    """Class to represent each balance sheet item in accounting.

    Each balance sheet item in accounting have the following properties:
        - account number: the unique IDentifier of this item
        - name: name of this aitem
        - origin: the origin of item of this item
        - balance sheet sub main group: each item must belong to a
            a group in balance sheet
        - balance sheet main group: this property is readonly
        and is determined by balance sheet sub main group

    """

    ID: str
    name: str
    origin: str | None = None
    balance_sheet_sub_main_group: BalanceSheetSubMainGroup

    @property
    def balance_sheet_main_group(self) -> BalanceSheetMainGroup:
        return self.balance_sheet_sub_main_group.balance_sheet_main_group

    @property
    def bsmain(self) -> BalanceSheetMainGroup:
        return self.balance_sheet_main_group

    @property
    def bssubmain(self) -> BalanceSheetSubMainGroup:
        return self.balance_sheet_sub_main_group

    def __str__(self) -> str:
        return str(self.ID) + " - " + self.name

    def to_csv(self, depth: int = 0, end: str = "\n") -> str:
        assert type(end) is str
        csv = ""
        if depth >= 0:
            csv += f"{self.ID},{self.name},{self.origin}"
        if depth >= 1:
            csv += ","
            csv += f"{self.balance_sheet_sub_main_group.to_csv(depth=depth-1)}"
        csv += end
        return csv

    def is_main_asset(self) -> bool:
        return self.bsmain is BSMain.ASSET

    def is_main_liability(self) -> bool:
        return self.bsmain is BSMain.LIABILITY

    def is_main_equity(self) -> bool:
        return self.bsmain is BSMain.EQUITY

    def is_short_term_asset(self) -> bool:
        return self.bssubmain is BSSubMain.SHORT_TERM_ASSET

    def is_long_term_asset(self) -> bool:
        return self.bssubmain is BSSubMain.LONG_TERM_ASSET

    def is_short_term_liability(self) -> bool:
        return self.bssubmain is BSSubMain.SHORT_TERM_LIABILITY

    def is_long_term_liability(self) -> bool:
        return self.bssubmain is BSSubMain.LONG_TERM_LIABILITY

    def is_equity(self) -> bool:
        return self.bssubmain is BSSubMain.OWNERS_EQUITY



from abc import ABC
from abc import abstractmethod
from typing import Type

from pydantic import BaseModel as PydanticBaseModel
from sqlalchemy.orm import DeclarativeBase as SqlAlchemyBaseModel
from sqlalchemy.orm import Session

from xarizmi.db.client import session_scope
from xarizmi.db.models.symbol import Symbol as SqlAlchemySymbol


class PydanticModelDbMapper(ABC):

    def __init__(
        self,
        pydantic_class: Type[PydanticBaseModel],
        sqlalchemy_class: Type[SqlAlchemyBaseModel],
    ) -> None:
        self._pydantic_class = pydantic_class
        self._sqlalchemy_class = sqlalchemy_class

    def _refresh(self, obj: PydanticBaseModel | SqlAlchemyBaseModel) -> None:
        if isinstance(obj, self._sqlalchemy_class):
            self._sqlalchemy_obj = obj
            self._pydantic_obj = self._to_pydantic(obj)
        elif isinstance(obj, self._pydantic_class):
            self._sqlalchemy_obj = self._to_sqlalchemy(obj)
            self._pydantic_obj = obj

    @abstractmethod
    def _to_pydantic(self, obj: SqlAlchemyBaseModel) -> PydanticBaseModel: ...

    @abstractmethod
    def _to_sqlalchemy(self, obj: PydanticBaseModel) -> SqlAlchemySymbol: ...

    @abstractmethod
    def _get_in_session(
        self, session: Session | None = None
    ) -> SqlAlchemySymbol | None: ...

    def get(
        self, session: None | Session = None
    ) -> SqlAlchemyBaseModel | None:
        if session is None:
            with session_scope() as session:
                record = self._get_in_session(session=session)
        else:
            record = self._get_in_session(session=session)
        return record

    def _does_exist_in_session(self, session: Session) -> bool:
        record = self._get_in_session(session=session)
        if record is None:
            return False
        else:
            return True



from datetime import datetime
from typing import Optional

from sqlalchemy import select
from sqlalchemy.orm import Session
from sqlalchemy.sql import and_

from xarizmi.candlestick import Candlestick as PyCandleStick
from xarizmi.candlestick import CandlestickChart as PyCandlestickChart
from xarizmi.db.actions.symbol import get_symbol
from xarizmi.db.models.candlestick import CandleStick
from xarizmi.db.models.symbol import Symbol
from xarizmi.enums import IntervalTypeEnum
from xarizmi.models.symbol import Symbol as PySymbol


def get_filtered_candlesticks(
    session: Session,
    symbol: Optional[PySymbol] = None,
    start_datetime: Optional[datetime] = None,
    end_datetime: Optional[datetime] = None,
    filter_by_interval_type: IntervalTypeEnum | None = None,
    skip: int = 0,
    limit: int = 1000,
) -> PyCandlestickChart:

    # Base query
    query = select(
        CandleStick,  # All columns from CandleStick
        Symbol.name.label("symbol_name"),  # Symbol name with alias
    ).join(
        Symbol, CandleStick.symbol_id == Symbol.id
    )  # Join with Symbol

    # Add filters
    filters = []
    if symbol:
        filters.append(
            Symbol.id == get_symbol(symbol=symbol, session=session).id
        )
    if start_datetime:
        filters.append(CandleStick.datetime >= start_datetime)
    if end_datetime:
        filters.append(CandleStick.datetime <= end_datetime)
    if filter_by_interval_type:
        filters.append(CandleStick.interval_type == filter_by_interval_type)

    if filters:
        query = query.where(and_(*filters))

    # Add ordering, skip, and limit for pagination
    query = (
        query.order_by(CandleStick.datetime.asc()).offset(skip).limit(limit)
    )

    # Execute the query
    result: list[CandleStick] = [
        item.CandleStick for item in session.execute(query).all()
    ]

    return PyCandlestickChart(
        candles=[
            PyCandleStick(
                open=db_candlestick.open,
                close=db_candlestick.close,
                low=db_candlestick.low,
                high=db_candlestick.high,
                volume=db_candlestick.volume,
                amount=db_candlestick.amount,
                interval_type=db_candlestick.interval_type,
                interval=db_candlestick.interval,
                symbol=symbol,
                datetime=db_candlestick.datetime,
                exchange=None,
            )
            for db_candlestick in result
        ]
    )



import numpy as np
from scipy.signal import argrelextrema  # type: ignore


def find_local_minima_indexes(arr: list[int | float]) -> list[int]:
    """
    Finds the indices of local minima indexes in a list using scipy.

    Parameters:
    arr (list of int/float): The list of values to find local minima in.

    Returns:
    list of int: Indices of local minima.
    """
    return argrelextrema(np.array(arr), np.less_equal)[0].tolist()  # type: ignore  # noqa: E501


def find_local_minima_values(arr: list[int | float]) -> list[int | float]:
    """
    Finds the values of local minima values in a list using scipy.

    Parameters:
    arr (list of int/float): The list of values to find local minima in.

    Returns:
    list of int/float: Values of local minima.
    """
    return [arr[i] for i in find_local_minima_indexes(arr)]


def find_local_maxima_indexes(arr: list[int | float]) -> list[int]:
    """
    Finds the indices of local maxima indexes in a list using scipy.

    Parameters:
    arr (list of int/float): The list of values to find local maxima in.

    Returns:
    list of int: Indices of local maxima.
    """
    return argrelextrema(np.array(arr), np.greater_equal)[0].tolist()  # type: ignore  # noqa: E501


def find_local_maxima_values(arr: list[int | float]) -> list[int | float]:
    """
    Finds the values of local maxima values in a list using scipy.

    Parameters:
    arr (list of int/float): The list of values to find local maxima in.

    Returns:
    list of int/float: Values of local maxima.
    """
    return [arr[i] for i in find_local_maxima_indexes(arr)]


def find_local_minima_of_minima_indexes(arr: list[int | float]) -> list[int]:
    """
    Finds the indices of local minima of the local minima in the original list.

    Parameters:
    arr (list of int/float): The list of values to find local minima in.

    Returns:
    list of int: Indices of local minima of the local minima in the original
    list.
    """
    # Find indices and values of local minima in the original list
    local_minima_indexes = find_local_minima_indexes(arr)
    local_minima_values = [arr[i] for i in local_minima_indexes]

    # Find indices of local minima in the list of local minima values
    local_minima_of_minima_indexes = find_local_minima_indexes(
        local_minima_values
    )

    # Map these indices back to the original list's indices
    return [local_minima_indexes[i] for i in local_minima_of_minima_indexes]


def find_local_maxima_of_maxima_indexes(arr: list[int | float]) -> list[int]:
    """
    Finds the indices of local maxima of the local maxima in the original list.

    Parameters:
    arr (list of int/float): The list of values to find local maxima in.

    Returns:
    list of int: Indices of local maxima of the local maxima in the original
    list.
    """
    # Find indices and values of local maxima in the original list
    local_maxima_indexes = find_local_maxima_indexes(arr)
    local_maxima_values = [arr[i] for i in local_maxima_indexes]

    # Find indices of local maxima in the list of local maxima values
    local_maxima_of_maxima_indexes = find_local_maxima_indexes(
        local_maxima_values
    )

    # Map these indices back to the original list's indices
    return [local_maxima_indexes[i] for i in local_maxima_of_maxima_indexes]



from sqlalchemy import Enum
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint

from xarizmi.db.models.base import Base
from xarizmi.enums import OrderStatusEnum
from xarizmi.enums import SideEnum

from .constants import TableNamesEnum
from .symbol import Symbol


class Order(Base):  # type: ignore
    __tablename__ = TableNamesEnum.ORDER.value
    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    symbol_id: Mapped[int] = mapped_column(
        Integer, ForeignKey(Symbol.id), nullable=False
    )
    symbol: Mapped[Symbol] = relationship("Symbol", back_populates="orders")

    order_id: Mapped[str] = mapped_column(String, nullable=False)
    amount: Mapped[float] = mapped_column(Float, nullable=False)
    price: Mapped[float] = mapped_column(Float, nullable=False)
    side: Mapped[str] = mapped_column(
        Enum(SideEnum, name="side_enum", create_type=True),
        nullable=False,
    )
    status: Mapped[str] = mapped_column(
        Enum(OrderStatusEnum, name="order_status_enum", create_type=True),
        nullable=False,
    )

    __table_args__ = (
        UniqueConstraint(
            "symbol_id",
            "order_id",
            name="uix_symbol_id_order_id",
        ),
    )



from .currency import Currency
from .exchange import Exchange
from .portfolio import Portfolio
from .portfolio import PortfolioItem
from .symbol import Symbol

__all__ = [
    "Portfolio",
    "PortfolioItem",
    "Symbol",
    "Currency",
    "Exchange",
]



from math import inf


def divide_with_no_zero_division_error(
    num: float | int, den: float | int
) -> float:
    try:
        return float(num) / float(den)
    except ZeroDivisionError:
        return inf



from enum import StrEnum


class TableNamesEnum(StrEnum):
    EXCHANGE = "xarizmi_exchange"
    PORTFOLIO_ITEM = "xarizmi_portfolio_item"
    SYMBOL = "xarizmi_symbol"
    CANDLESTICK = "xarizmi_candlestick"
    ORDER = "xarizmi_order"



from .portfolio_create import bulk_upsert_portfolio_item
from .portfolio_create import upsert_portfolio
from .portfolio_create import upsert_portfolio_item
from .portfolio_read import get_portfolio_items_between_dates

__all__ = [
    "upsert_portfolio",
    "upsert_portfolio_item",
    "bulk_upsert_portfolio_item",
    "get_portfolio_items_between_dates",
]



from pydantic import BaseModel

from xarizmi.enums import OrderStatusEnum
from xarizmi.enums import SideEnum

from .symbol import Symbol


class Order(BaseModel):
    symbol: Symbol
    price: float
    amount: float
    status: OrderStatusEnum
    side: SideEnum
    # order_id only exists when order is put in place in a trading platform
    # but to work with orders you don't need to have order_id necessarily
    order_id: str | None = None

    @classmethod
    def build_from_currencies(
        cls,
        base_currency: str,
        quote_currency: str,
        fee_currency: str,
        order_id: str,
        price: float,
        amount: float,
        status: OrderStatusEnum,
        side: SideEnum,
        exchange: str | None = None,
    ) -> "Order":

        symbol = Symbol.build(
            base_currency=base_currency,
            quote_currency=quote_currency,
            fee_currency=fee_currency,
            exchange=exchange,
        )
        data = {
            "symbol": symbol,
            "order_id": order_id,
            "price": price,
            "amount": amount,
            "side": side,
            "status": status,
        }
        return cls(**data)



from typing import Self

from pydantic import BaseModel

from xarizmi.models.currency import Currency
from xarizmi.models.exchange import Exchange


class Symbol(BaseModel):
    base_currency: Currency
    quote_currency: Currency
    fee_currency: Currency
    exchange: Exchange | None = None

    def __hash__(self) -> int:
        return hash((self.base_currency.name, self.quote_currency.name))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Symbol):
            raise NotImplementedError(
                f"Can't compare objects of type {type(other)} and {type(self)}"
            )
        return (
            self.base_currency == other.base_currency
            and self.quote_currency == other.quote_currency
            and self.fee_currency == other.fee_currency
            and self.exchange == other.exchange
        )

    @classmethod
    def build(
        cls,
        base_currency: str,
        quote_currency: str,
        fee_currency: str,
        exchange: str | None = None,
    ) -> "Symbol":
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        """
        data = {
            "base_currency": {"name": base_currency},
            "quote_currency": {"name": quote_currency},
            "fee_currency": {"name": fee_currency},
        }
        if exchange is not None:
            data["exchange"] = {"name": exchange}
        return cls(**data)

    def to_string(self) -> str:
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        >>> symbol.to_string()
        'BTC-USD'
        """
        return (
            self.base_currency.to_string()
            + "-"
            + self.quote_currency.to_string()
        )

    def to_flat_csv(self) -> str:
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        >>> symbol.to_string()
        'BTC-USD'
        """
        return (
            self.base_currency.to_string()
            + "-"
            + self.quote_currency.to_string()
        )

    def to_dict(self) -> dict[str, str | None]:
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        >>> symbol.to_dict()
        {
            "base_currency": "BTC",
            "quote_currency": "USD",
            "fee_currency": "USD"
            "exchange": "BINANCE",
        }
        """
        return {
            "base_currency": self.base_currency.to_string(),
            "quote_currency": self.quote_currency.to_string(),
            "fee_currency": self.fee_currency.to_string(),
            "exchange": (
                self.exchange.to_string()
                if (self.exchange is not None)
                else None
            ),
        }


class SymbolList(BaseModel):
    items: list[Symbol] = []

    @classmethod
    def build(
        cls,
        items: list[dict[str, str]],
    ) -> Self:
        symbol_items = []
        for item in items:
            symbol_items.append(
                Symbol.build(
                    base_currency=item["base_currency"],
                    quote_currency=item["quote_currency"],
                    fee_currency=item["fee_currency"],
                    exchange=item["exchange"],
                )
            )
        return cls(items=symbol_items)



from pathlib import Path

from alembic import command
from alembic.config import Config

from xarizmi.config import get_config

PARENT_DIR = Path(__file__).parent
alembic_ini = PARENT_DIR / "alembic.ini"


def run_db_migration() -> None:
    # Path to alembic.ini file
    alembic_cfg = Config(alembic_ini)

    # Dynamically set the database URL in Alembic's config
    alembic_cfg.set_main_option("sqlalchemy.url", get_config().DATABASE_URL)
    alembic_cfg.set_main_option(
        "script_location", (PARENT_DIR / "alembic").__str__()
    )

    command.upgrade(alembic_cfg, "head")



from typing import Self

from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import relationship

from xarizmi.db.models.base import Base
from xarizmi.models.exchange import Exchange as PyExchange

from .constants import TableNamesEnum


class Exchange(Base):  # type: ignore
    __tablename__ = TableNamesEnum.EXCHANGE.value
    name = Column(String, primary_key=True, unique=True)

    symbols: Mapped[list["Symbol"]] = relationship(  # type: ignore  # noqa: F821,E501
        "Symbol", back_populates="exchange"
    )

    def to_pydantic(self) -> PyExchange:
        return PyExchange(name=self.name)

    @classmethod
    def from_pydantic(cls, exchange: PyExchange) -> Self:
        return cls(name=exchange)



"""Module to calculate Present Value of Perpetuity
"""


def perpetuity_present_value(
    continuous_cash_payment: float,
    interest_rate: float,
) -> float:
    """Returns the Present Value of Perpetuity Formula.

    Parameters
    ----------
    continuous_cash_payment : float
        Amount of continuous cash payment.

    interest_rate : float
        Interest rate, yield or discount rate.

    Returns
    -------
    float:
        Present value of perpetuity

    Example
    -------
    >>> pv = perpetuity_present_value(5, 0.15)
    >>> round(pv, 2)
    33.33
    """
    return continuous_cash_payment / interest_rate


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



"""A module to calculate cash flow
"""

from pydantic import BaseModel
from pydantic import PositiveInt


def compound_future(
    present_value: float, interest_rate: float, n_years: int
) -> float:
    return present_value * (1 + interest_rate) ** n_years


def compound_present(
    future_value: float, interest_rate: float, n_years: int
) -> float:
    return future_value / (1 + interest_rate) ** n_years


class AnnualCashFlowItem(BaseModel):
    """
    Example
    -------
    >>> ci = AnnualCashFlowItem(0, -1000, 'operating cost', 'perpetuity')
    >>> print(ci)
    0 ----------   -1000 ---------- operating cost            perpetuity
    """

    year: PositiveInt
    cf: float
    event: str | None = None
    kind: str | None = (
        None  # TODO: should be enum with "perpetuity" as one of values # noqa: E501
    )

    def __str__(self) -> str:
        s = ""
        s += (f"{self.year} " + "-" * 10).ljust(10) + " "
        s += str(self.cf).rjust(len(str(self.cf)) + 2)
        if self.event:
            s += " " + ("-" * 10 + " " + self.event).rjust(25)
        if self.kind:
            s += " " + (" " * 10 + " " + self.kind).rjust(10)
        return s


class AnnualCashFlow:
    """
    Example:
    >>> cashflow = AnnualCashFlow(
    [(0, -2000), (1, 600), (2, 600), (3, 600), (4, 600)], 0.1
    )
    >>> print(cashflow)
    ============ CASH FLOW ============
    0 ------------------------- -2000
    1 -------------------------   600
    2 -------------------------   600
    3 -------------------------   600
    4 -------------------------   600
    ____________________________________________________________
    Total                        400          INFLOW          2400
                                              OUTFLOW         2000
    ======= DISCOUNTED CASH FLOW =======
    0 -------------------------  -2000.00
    1 -------------------------    545.45
    2 -------------------------    495.87
    3 -------------------------    450.79
    4 -------------------------    409.81
    ______________________________________________________________
    Total                       -98.08          INFLOW       1901.92
                                                OUTFLOW       2000.0
    >>> cashflow.calculate_discounted_cash_flow(rounding=2)
    [
        (0, -2000.0, None),
        (1, 545.45, None),
        (2, 495.87, None),
        (3, 450.79, None),
        (4, 409.81, None),
    ]
    """

    def __init__(
        self,
        cash_flow: (
            list[tuple[int, float]] | list[tuple[int, float, None | str]]
        ),
        time_value_money: float = 0,
    ) -> None:
        self.cash_flow = cash_flow  # type: ignore
        self.time_value_money = time_value_money
        self.discounted_cash_flow = self.calculate_discounted_cash_flow()

    @property
    def cash_flow(
        self,
    ) -> list[tuple[int, float, None | str]]:
        return self._cash_flow

    @cash_flow.setter
    def cash_flow(
        self,
        cash_flow: (
            list[tuple[int, float]] | list[tuple[int, float, str | None]]
        ),
    ) -> None:
        assert type(cash_flow) is list
        for item in cash_flow:
            assert type(item) is tuple
            assert len(item) == 2 or len(item) == 3
        self._cash_flow: list[tuple[int, float, str | None]] = []
        for item in cash_flow:
            if len(item) == 2:
                self._cash_flow.append((item[0], item[1], None))
            elif len(item) == 3:
                self._cash_flow.append((item[0], item[1], item[2]))
            else:
                raise NotImplementedError()

    def netflow(self) -> float:
        return sum(cf for _, cf, _ in self.cash_flow)

    def inflow(self) -> float:
        return sum(
            filter(lambda cf: cf > 0, (cf for _, cf, _ in self._cash_flow))
        )

    def outflow(self) -> float:
        return abs(
            sum(
                filter(lambda cf: cf < 0, (cf for _, cf, _ in self._cash_flow))
            )
        )

    def discounted_netflow(self) -> float:
        return sum(cf for _, cf, _ in self.discounted_cash_flow)

    def discounted_inflow(self) -> float:
        return sum(
            filter(
                lambda cf: cf > 0,
                (cf for _, cf, _ in self.discounted_cash_flow),
            )
        )

    def discounted_outflow(self) -> float:
        return abs(
            sum(
                filter(
                    lambda cf: cf < 0,
                    (cf for _, cf, _ in self.discounted_cash_flow),
                )
            )
        )

    def __str__(self) -> str:
        s = "============ CASH FLOW ============\n"
        max_length = max(len(str(item[1])) for item in self.cash_flow)
        for year, cf, event in self.cash_flow:
            s += (f"{year} " + "-" * 25).ljust(25) + " "
            s += str(cf).rjust(max_length)
            if event:
                s += " " + "-" * 10 + " " + event.rjust(20)
            s += "\n"
        s += "_" * (25 + max_length + 10 + 20) + "\n"
        s += (
            ("Total" + " " * 22).ljust(25)
            + str(self.netflow()).rjust(max_length)
            + " " * 10
            + "INFLOW".ljust(10)
            + str(self.inflow()).rjust(10)
            + "\n"
        )
        s += (
            (" " * (5 + 22 + max_length))
            + " " * 10
            + "OUTFLOW".ljust(10)
            + str(self.outflow()).rjust(10)
            + "\n"
        )
        if self.time_value_money:
            s += "======= DISCOUNTED CASH FLOW =======\n"
            discounted = self.calculate_discounted_cash_flow(rounding=2)
            for year, cf, event in discounted:
                max_length = max(len(str(item[1])) for item in discounted)
                s += (f"{year} " + "-" * 25).ljust(25) + " "
                s += ("{:.2f}".format(cf)).rjust(max_length + 2)
                if event:
                    s += " " + "-" * 10 + " " + event.rjust(20)
                s += "\n"
        s += "_" * (25 + max_length + 10 + 20) + "\n"
        s += (
            ("Total" + " " * 22).ljust(25)
            + str(round(self.discounted_netflow(), 2)).rjust(max_length)
            + " " * 10
            + "INFLOW".ljust(10)
            + str(round(self.discounted_inflow(), 2)).rjust(10)
            + "\n"
        )
        s += (
            (" " * (5 + 22 + max_length))
            + " " * 10
            + "OUTFLOW".ljust(10)
            + str(round(self.discounted_outflow(), 2)).rjust(10)
            + "\n"
        )
        return s[:-1]

    def calculate_discounted_cash_flow(
        self, rounding: int | None = None
    ) -> list[tuple[int, float, None | str]]:
        self.discounted_cash_flow = []
        for year, future_cf, event in self.cash_flow:
            discounted = compound_present(
                future_cf, self.time_value_money, year
            )
            if rounding is None:
                pass
            else:
                assert type(rounding) is int and rounding >= 0
                discounted = round(discounted, rounding)
            self.discounted_cash_flow.append((year, discounted, event))
        return self.discounted_cash_flow

    def npv(self) -> float:
        """
        Example:
        >>> cashflow = AnnualCashFlow([(0, -5), (0, -6), (1, 12)], 0.15)
        >>> npv = cashflow.npv()
        >>> round(npv, 3)
        -0.565
        """
        return sum(cf for _, cf, _ in self.calculate_discounted_cash_flow())


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



from .export import export_all_xarizmi_tables_to_json
from .export import export_table_to_csv
from .export import export_table_to_json
from .load import (
    load_pydantic_candlestick_model_from_db_model_exported_json_file,
)
from .load import load_pydantic_exchange_model_from_db_model_exported_json_file
from .load import load_pydantic_symbol_model_from_db_model_exported_json_file

__all__ = [
    "export_table_to_csv",
    "export_table_to_json",
    "export_all_xarizmi_tables_to_json",
    "load_pydantic_candlestick_model_from_db_model_exported_json_file",
    "load_pydantic_symbol_model_from_db_model_exported_json_file",
    "load_pydantic_exchange_model_from_db_model_exported_json_file",
]



"""Module for Financial Transactions
"""

import datetime


class Transaction:

    def __init__(
        self,
        ID: str,
        amount: float,
        inflow: bool,
        time: datetime.datetime,
        description: str = "",
        currency_code: str = "USD",
    ) -> None:
        self.ID = ID
        self.amount = amount
        self.inflow = inflow
        self.time = time
        self.description = description
        self.currency_code = currency_code

    @property
    def amount(self) -> float:
        return self._amount

    @amount.setter
    def amount(self, amount: float | int) -> None:
        assert type(amount) is float or type(amount) is int
        assert amount > 0
        self._amount = amount

    @property
    def inflow(self) -> bool:
        return self._inflow

    @inflow.setter
    def inflow(self, inflow: int | float) -> None:
        assert type(inflow) is bool
        self._inflow = inflow

    @property
    def time(self) -> datetime.datetime:
        return self._time

    @time.setter
    def time(self, time: datetime.datetime) -> None:
        assert type(time) is datetime.datetime
        self._time = time
        self._date = time.date()

    @property
    def date(self) -> datetime.date:
        return self._date

    @property
    def currency_code(self) -> str:
        return self._currency_code

    @currency_code.setter
    def currency_code(self, currency_code: str) -> None:
        assert type(currency_code) is str
        assert currency_code in [
            "USD",
            "IRR",
            "AUD",
            "EUR",
        ]
        self._currency_code = currency_code

    def __str__(self) -> str:
        return (
            f"{'+' if self.inflow else '-'}{self.amount:,}"
            " - {self.description}"
        )



from pydantic import BaseModel
from pydantic import model_validator


class Fundamentals(BaseModel):
    """
    reported_earnings: float
    Reported earnings refer to the net income
    disclosed in financial statements as per regulatory standards
    (e.g., GAAP or IFRS). It reflects what the company officially
    "reports" to stakeholders.

    number_of_shares: int
    number of outstanding shares of the company.
    Treasury shares are shared bought by company and kept in
    treasury are not part of this.

    enterprise_value: float
    When you buy a company you own its “good things”
    (equity) and “bad things” (debt). The amount of “theoretical”
    money you need to buy a company and pays all its debit is called
    Enterprise value.


    """

    reported_earnings: float
    number_of_shares: int
    share_price: float
    reported_earnings_before_interest_tax_depreciation_amortization: float
    cash_and_cash_equivalents: float
    total_debt: float
    enterprise_value: float

    @model_validator(mode="before")
    def handle_enterprise_value(
        cls, values: dict[str, int | float]
    ) -> dict[str, int | float]:
        if "enterprise_value" not in values:
            values["enterprise_value"] = (
                values["number_of_shares"] * values["share_price"]
                + values["total_debt"]
                - values["cash_and_cash_equivalent"]
            )
        return values

    @property
    def market_capitalization(self) -> float:
        """Market Capitalization: The total value of a
        company's outstanding shares of stock.
        It's calculated by multiplying the current share price
        by the total number of outstanding shares.

        The `shares` attribute in Fundamentals is actually
        the "outstanding shares" of the company.
        """
        return self.number_of_shares * self.share_price

    @property
    def EBITDA(self) -> float:
        return (
            self.reported_earnings_before_interest_tax_depreciation_amortization  # noqa: E501
        )

    @property
    def enterprise_value_to_EBITDA_ratio(self) -> float:
        return (
            self.enterprise_value
            / self.reported_earnings_before_interest_tax_depreciation_amortization  # noqa: E501
        )

    @property
    def earnings_per_share(self) -> float:
        return self.reported_earnings / self.number_of_shares

    @property
    def EPS(self) -> float:
        """Returns earnings per share"""
        return self.earnings_per_share

    @property
    def price_to_earnings_ratio(self) -> float:
        return self.share_price / self.earnings_per_share

    @property
    def PE(self) -> float:
        """Returns P/E (price to earnings per share)"""
        return self.price_to_earnings_ratio



from datetime import datetime as dt
from typing import Any

import mplfinance as mpf
import numpy as np
import pandas as pd
from pydantic import BaseModel
from pydantic import NonNegativeFloat

from xarizmi.config import config
from xarizmi.enums import IntervalTypeEnum
from xarizmi.models.exchange import Exchange
from xarizmi.models.symbol import Symbol
from xarizmi.utils.extremums import find_local_maxima_indexes
from xarizmi.utils.extremums import find_local_maxima_of_maxima_indexes
from xarizmi.utils.extremums import find_local_minima_indexes
from xarizmi.utils.extremums import find_local_minima_of_minima_indexes
from xarizmi.utils.extremums import find_local_minima_values
from xarizmi.utils.numbers import round_to_significant_digit


class Candlestick(BaseModel):
    """This class contains basic properties of a candlestick.

    :param close: the closing price for the corresponding time period
    :type close: float
    :param open: the opening price for the corresponding time period
    :type open: float
    :param low: the lowest traded price for the corresponding time period
    :type low: float
    :param high: the highest traded price for the corresponding time period
    :type high: float
    :param volume: the amount of trading volume (width of the candlestick)
    :type volume: float
    :param amount: [amount description?]
    :type amount: float
    :param interval_type: the time period for each candlestick
    :type interval_type: str
    :param symbol: the type of the trade, e.g., "CAD-USD" or "BTC-USDT"
    :type symbol: str
    """

    close: NonNegativeFloat
    open: NonNegativeFloat
    low: NonNegativeFloat
    high: NonNegativeFloat
    volume: NonNegativeFloat
    amount: NonNegativeFloat | None = None
    interval_type: IntervalTypeEnum | None = None
    interval: int | None = None  # interval in seconds
    symbol: Symbol | None = None
    datetime: dt | None = None
    exchange: Exchange | None = None

    @property
    def is_bullish(self) -> bool:
        if self.close >= self.open:
            return True
        else:
            return False

    @property
    def is_bearish(self) -> bool:
        if self.open >= self.close:
            return True
        else:
            return False

    @property
    def range(self) -> float:
        """Range = H - L"""
        return self.high - self.low

    @property
    def intrinsic_range(self) -> float:
        """
        IR = R / L
        """
        if (self.low) == 0:
            return 0
        else:
            return self.range / (self.low)

    @property
    def body(self) -> float:
        """B = O - C"""
        return abs(self.open - self.close)

    @property
    def intrinsic_body(self) -> float:
        """B / R"""
        if self.range == 0:
            return 0
        return self.body / self.range

    @property
    def upper_shadow(self) -> float:
        """US = H - MAX(C, O)"""
        return self.high - max(self.close, self.open)

    @property
    def intrinsic_upper_shadow(self) -> float:
        """US / R"""
        if self.range == 0:
            return 0
        return self.upper_shadow / self.range

    @property
    def lower_shadow(self) -> float:
        """LS = min(C, O) - L"""
        return min(self.close, self.open) - self.low

    @property
    def intrinsic_lower_shadow(self) -> float:
        """LS / R"""
        if self.range == 0:
            return 0
        return self.lower_shadow / self.range

    @property
    def doginess(self) -> float:
        if self.range == 0:
            return 0
        return 1 - self.intrinsic_body

    @property
    def is_doji(self) -> bool:
        if self.range == 0:
            return False
        return self.doginess >= config.DOJINESS_THRESHOLD

    def flatten(self) -> dict[str, Any]:
        if self.symbol is None:
            raise NotImplementedError(
                "flatten method is not implemented for instances with "
                "None value for self.symbol"
            )
        data = self.model_dump()
        del data["symbol"]
        data["base_currency"] = self.symbol.base_currency.name
        data["quote_currency"] = self.symbol.quote_currency.name
        data["exchange"] = (
            self.exchange.name if (self.exchange is not None) else None
        )
        return data


class CandlestickChart(BaseModel):
    candles: list[Candlestick]

    def filter_keep_last_n(
        self, keep_last_n: int | None = None, inplace: bool = False
    ) -> "CandlestickChart":
        if keep_last_n:
            candles = self.candles[-keep_last_n:]
            if inplace is True:
                self.candles = candles
                return self
            else:
                return CandlestickChart(candles=candles)
        return self

    def to_df(self) -> pd.DataFrame:
        df = pd.DataFrame(
            data={
                "open": [candle.open for candle in self.candles],
                "close": [candle.close for candle in self.candles],
                "low": [candle.low for candle in self.candles],
                "high": [candle.high for candle in self.candles],
                "datetime": [candle.datetime for candle in self.candles],
                "volume": [candle.volume for candle in self.candles],
                "amount": [candle.amount for candle in self.candles],
            }
        )
        df.set_index("datetime", inplace=True)
        return df

    def save_simple_plot_no_volume(self, filepath: str) -> None:
        custom_style = mpf.make_mpf_style(  # type: ignore
            base_mpf_style="classic",
            marketcolors=mpf.make_marketcolors(  # type: ignore
                up="green",
                down="red",
                wick="black",
                edge="inherit",
            ),
        )
        mpf.plot(
            self.to_df(),
            type="candle",
            style=custom_style,
            title="Candlestick Chart",
            ylabel="Price",
            savefig="candlestick_chart.svg",
        )

    def plot(self, figsize: tuple[int | float, int | float] = (8, 4)) -> None:

        df = self.to_df()

        mask = np.isin(df.index, df.index[self.get_support_indexes()])
        masked_df = df.copy()
        masked_df[~mask] = np.nan  # Replace rows not in index_list with NaN
        apd1 = mpf.make_addplot(
            masked_df["low"],
            type="scatter",
            color="darkgreen",
            marker="h",
            markersize=20,
            alpha=0.50,
        )
        mask = np.isin(df.index, df.index[self.get_l2_support_indexes()])
        masked_df = df.copy()
        masked_df[~mask] = np.nan  # Replace rows not in index_list with NaN
        apd3 = mpf.make_addplot(
            masked_df["low"],
            type="scatter",
            color="darkgreen",
            marker="s",
            markersize=100,
            alpha=0.5,
        )
        mask = np.isin(df.index, df.index[self.get_resistance_indexes()])
        masked_df = df.copy()
        masked_df[~mask] = np.nan  # Replace rows not in index_list with NaN
        apd2 = mpf.make_addplot(
            masked_df["high"],
            type="scatter",
            color="darkred",
            marker="h",
            markersize=20,
            alpha=0.50,
        )
        mask = np.isin(df.index, df.index[self.get_l2_resistance_indexes()])
        masked_df = df.copy()
        masked_df[~mask] = np.nan  # Replace rows not in index_list with NaN
        apd4 = mpf.make_addplot(
            masked_df["high"],
            type="scatter",
            color="darkred",
            marker="s",
            markersize=100,
            alpha=0.5,
        )
        mpf.plot(
            df,
            figratio=figsize,
            type="candle",
            volume=True,
            # style="yahoo",
            style="classic",
            addplot=[apd1, apd2, apd3, apd4],
        )

    def get_local_minimas(
        self, price_type: str = "low", only_significant_digit: bool = False
    ) -> list[int | float]:
        if price_type not in ["low", "high", "close", "open"]:
            raise ValueError(
                f"The given value for {price_type=}" f" is not valid!"
            )
        values = find_local_minima_values(
            [getattr(candle, price_type) for candle in self.candles]
        )
        if only_significant_digit is True:
            values = [round_to_significant_digit(item) for item in values]
        return values

    def get_support_indexes(self) -> list[int]:
        return find_local_minima_indexes(
            [candle.low for candle in self.candles]
        )

    def get_l2_support_indexes(self) -> list[int]:
        return find_local_minima_of_minima_indexes(
            [candle.low for candle in self.candles]
        )

    def get_resistance_indexes(self) -> list[int]:
        return find_local_maxima_indexes(
            [candle.high for candle in self.candles]
        )

    def get_l2_resistance_indexes(self) -> list[int]:
        return find_local_maxima_of_maxima_indexes(
            [candle.high for candle in self.candles]
        )


class CandlestickList(BaseModel):
    items: list[Candlestick] = []



from .balance_sheet_item import BalanceSheetItem


class BalanceSheetItemTuple:

    def __init__(
        self,
        balance_sheet_item: BalanceSheetItem,
        debit: float,
        credit: float,
    ) -> None:
        self.balance_sheet_item = balance_sheet_item
        self.debit = debit
        self.credit = credit

    @property
    def balance_sheet_item(self) -> BalanceSheetItem:
        return self._balance_sheet_item

    @balance_sheet_item.setter
    def balance_sheet_item(self, balance_sheet_item: BalanceSheetItem) -> None:
        assert isinstance(balance_sheet_item, BalanceSheetItem)
        self._balance_sheet_item = balance_sheet_item



import datetime
from datetime import datetime as dt

import pytz

from xarizmi.candlestick import Candlestick
from xarizmi.config import get_config
from xarizmi.db import run_db_migration
from xarizmi.db.actions.candlestick import get_filtered_candlesticks
from xarizmi.db.actions.candlestick import upsert_candlestick
from xarizmi.db.actions.exchange import bulk_upsert_exchanges
from xarizmi.db.actions.order import delete_all_cancelled_orders
from xarizmi.db.actions.order import delete_unique_order
from xarizmi.db.actions.order import get_orders
from xarizmi.db.actions.order import get_unique_order
from xarizmi.db.actions.order import upsert_order
from xarizmi.db.actions.portfolio import upsert_portfolio
from xarizmi.db.actions.portfolio.portfolio_read import (
    get_portfolio_items_between_dates,
)
from xarizmi.db.actions.symbol import bulk_upsert_symbols
from xarizmi.db.actions.symbol import get_symbol
from xarizmi.db.client import session_scope
from xarizmi.enums import IntervalTypeEnum
from xarizmi.enums import OrderStatusEnum
from xarizmi.enums import SideEnum
from xarizmi.models.exchange import Exchange
from xarizmi.models.orders import Order
from xarizmi.models.portfolio import Portfolio
from xarizmi.models.portfolio import PortfolioItem
from xarizmi.models.symbol import Symbol


def xarizmi_db_example() -> None:
    config = get_config()
    config.DATABASE_URL = "postgresql://postgres:1@localhost/xarizmi"

    run_db_migration()
    # insert exchanges to exchange table
    with session_scope() as session:
        exchanges = bulk_upsert_exchanges(
            exchanges=[
                Exchange(name=name)
                for name in [
                    "KUCOIN",
                    "crypto.com",
                    "BINANCE",
                    "COINBASE",
                ]
            ],
            session=session,
        )
        print(
            "Exchanges created in db:\n "
            f"{[vars(exchange) for exchange in exchanges]}",
            end="\n-----------\n",
        )

    # insert symbols to Symbol table
    with session_scope() as session:
        bulk_upsert_symbols(
            [
                Symbol.build(
                    base_currency="BTC",
                    quote_currency="USDT",
                    fee_currency="USDT",
                    exchange="COINBASE",
                ),
                Symbol.build(
                    base_currency="ETH",
                    quote_currency="USDT",
                    fee_currency="USDT",
                    exchange="BINANCE",
                ),
                Symbol.build(
                    base_currency="CRO",
                    quote_currency="USDT",
                    fee_currency="USDT",
                    exchange="KUCOIN",
                ),
                Symbol.build(
                    base_currency="CRO",
                    quote_currency="USD",
                    fee_currency="USD",
                    exchange="crypto.com",
                ),
                Symbol.build(
                    base_currency="BTC",
                    quote_currency="USD",
                    fee_currency="USD",
                    exchange="crypto.com",
                ),
            ],
            session=session,
        )

    target_symbol = Symbol.build(
        base_currency="CRO",
        quote_currency="USD",
        fee_currency="USD",
        exchange="crypto.com",
    )

    symbol_id = get_symbol(target_symbol, session=session).id
    upsert_candlestick(
        symbol_id=symbol_id,
        session=session,
        candlestick=Candlestick(
            symbol=target_symbol,
            close=1000,
            open=4,
            low=3,
            high=30000,
            volume=10000,
            amount=1,
            interval=1732385697000,
            datetime=dt(2024, 11, 23, 18, 14, 0, tzinfo=pytz.UTC),
            interval_type=IntervalTypeEnum.HOUR_1,
        ),
    )

    candlestick_chart = get_filtered_candlesticks(
        session=session,
        symbol=target_symbol,
        filter_by_interval_type=IntervalTypeEnum.DAY_14,
    )
    candlestick_chart.plot()
    print(">>>>>>>>>>>>>>>>>>>>>")

    portfolio = Portfolio(
        items=[
            PortfolioItem(
                symbol=Symbol.build(
                    base_currency="CRO",
                    quote_currency="USD",
                    fee_currency="USD",
                    exchange="crypto.com",
                ),
                market_value=1000,
                quantity=0.001,
                datetime=dt(2024, 11, 25),
            ),
            PortfolioItem(
                symbol=Symbol.build(
                    base_currency="BTC",
                    quote_currency="USD",
                    fee_currency="USD",
                    exchange="crypto.com",
                ),
                market_value=1000,
                quantity=10000,
                datetime=dt(2024, 11, 25),
            ),
            PortfolioItem(
                symbol=Symbol.build(
                    base_currency="BTC",
                    quote_currency="USD",
                    fee_currency="USD",
                    exchange="crypto.com",
                ),
                market_value=2000,
                quantity=10000,
                datetime=dt(2024, 11, 25),
            ),
        ]
    )

    upsert_portfolio(
        portfolio=portfolio,
        session=session,
    )

    print(result)

    res = get_portfolio_items_between_dates(
        session=session,
        start_date=datetime.datetime(2024, 11, 25),
        end_date=datetime.datetime(2024, 11, 25),
    )
    print(res)

    # ============== Order ===================
    BTC_USDT = Symbol.build(
        base_currency="BTC",
        quote_currency="USDT",
        fee_currency="USDT",
        exchange="COINBASE",
    )
    # create pydantic order
    order = Order(
        symbol=BTC_USDT,
        price=75000,
        amount=0.1,
        status=OrderStatusEnum.ACTIVE,
        side=SideEnum.BUY,
        order_id="FAKEORDERID",
    )
    db_order = upsert_order(order=order, session=session)
    # an other order
    order = Order(
        symbol=BTC_USDT,
        price=50000,
        amount=0.2,
        status=OrderStatusEnum.DONE,
        side=SideEnum.SELL,
        order_id="FAKEORDERID_2",
    )
    db_order = upsert_order(order=order, session=session)
    # an other order with CANCELLED status
    order = Order(
        symbol=BTC_USDT,
        price=1000000,
        amount=0.2,
        status=OrderStatusEnum.CANCELLED,
        side=SideEnum.BUY,
        order_id="FAKEORDERID_3",
    )
    db_order = upsert_order(order=order, session=session)

    order = get_unique_order(
        session=session,
        symbol=BTC_USDT,
        order_id="FAKEORDERID",
    )
    print(f"Pydantic Order read from DB: {order=}")

    # get all orders
    orders = get_orders(
        session=session,
        symbol=BTC_USDT,
    )
    print(f"Number of all orders in order table: {len(orders)}")
    # get active buy orders
    orders = get_orders(
        session=session,
        symbol=BTC_USDT,
        filter_by_side=SideEnum.BUY,
        filter_by_status=[OrderStatusEnum.ACTIVE],
    )
    print(f"Number of all active orders in order table: {len(orders)}")

    delete_unique_order(
        session=session, symbol=BTC_USDT, order_id="FAKEORDERID_2"
    )
    # get all orders after delete on record
    orders = get_orders(
        session=session,
        symbol=BTC_USDT,
    )
    print(f"Number of all orders in order table: {len(orders)}")

    # delete all cancelled orders
    delete_all_cancelled_orders(session=session)
    # get all orders after delete on record
    orders = get_orders(
        session=session,
        symbol=BTC_USDT,
    )
    print(f"Number of all orders in order table: {len(orders)}")


# export_table_to_csv("xarizmi_symbol", "xarizmi_symbol.csv", get_engine())



"""add_more_fields

Revision ID: 0f106fabd22f
Revises: 4bcb140de81f
Create Date: 2024-11-23 10:33:27.024696

"""

from typing import Sequence
from typing import Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import ENUM

# revision identifiers, used by Alembic.
revision: str = "0f106fabd22f"
down_revision: Union[str, None] = "4bcb140de81f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # create enum type
    # Create the enum type
    interval_type_enum = ENUM(
        "MIN_1",
        "MIN_3",
        "MIN_5",
        "MIN_15",
        "MIN_30",
        "HOUR_1",
        "HOUR_2",
        "HOUR_4",
        "HOUR_6",
        "HOUR_8",
        "HOUR_12",
        "DAY_1",
        "DAY_7",
        "DAY_14",
        "WEEK_1",
        "MONTH_1",
        name="interval_type_enum",
    )
    interval_type_enum.create(op.get_bind())  # type: ignore

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "xarizmi_candlestick", sa.Column("close", sa.Integer(), nullable=False)
    )
    op.add_column(
        "xarizmi_candlestick", sa.Column("open", sa.Integer(), nullable=False)
    )
    op.add_column(
        "xarizmi_candlestick", sa.Column("high", sa.Integer(), nullable=False)
    )
    op.add_column(
        "xarizmi_candlestick", sa.Column("low", sa.Integer(), nullable=False)
    )
    op.add_column(
        "xarizmi_candlestick",
        sa.Column("volume", sa.Integer(), nullable=False),
    )
    op.add_column(
        "xarizmi_candlestick", sa.Column("amount", sa.Integer(), nullable=True)
    )
    op.add_column(
        "xarizmi_candlestick",
        sa.Column("interval_type", interval_type_enum, nullable=False),
    )
    op.add_column(
        "xarizmi_candlestick",
        sa.Column("interval", sa.BigInteger(), nullable=True),
    )
    op.add_column(
        "xarizmi_candlestick",
        sa.Column("datetime", sa.DateTime(), nullable=False),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("xarizmi_candlestick", "datetime")
    op.drop_column("xarizmi_candlestick", "interval")
    op.drop_column("xarizmi_candlestick", "interval_type")
    op.drop_column("xarizmi_candlestick", "amount")
    op.drop_column("xarizmi_candlestick", "volume")
    op.drop_column("xarizmi_candlestick", "low")
    op.drop_column("xarizmi_candlestick", "high")
    op.drop_column("xarizmi_candlestick", "open")
    op.drop_column("xarizmi_candlestick", "close")
    # ### end Alembic commands ###

    # Drop the enum type
    interval_type_enum = ENUM(
        "MIN_1",
        "MIN_3",
        "MIN_5",
        "MIN_15",
        "MIN_30",
        "HOUR_1",
        "HOUR_2",
        "HOUR_4",
        "HOUR_6",
        "HOUR_8",
        "HOUR_12",
        "DAY_1",
        "DAY_7",
        "DAY_14",
        "WEEK_1",
        "MONTH_1",
        name="interval_type_enum",
    )
    interval_type_enum.drop(op.get_bind())  # type: ignore



"""
Module for main group of items in Balance Sheet.

There are Three main groups in a balance sheet:

    - Asset
    - Liability
    - Equity
"""

from pydantic import BaseModel


class BalanceSheetMainGroup(BaseModel):
    """
    Class to determine Balance Sheet Group.

    Balance Sheet Main Group is one the followings:
    - Asset
    - Liability
    - Equity

    Example
    -------
    >>> ASSET = BalanceSheetMainGroup(id=1, name="Asset")
    """

    id: int
    name: str

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"BalanceSheetMainGroup(id={self.id}, name={self.name})"

    def to_csv(self, depth: int = 0, end: str = "") -> str:
        """Returns a string which can be used to generated csv files.

        Parameters
        ----------
        depth : int
            If non-negative will return information of BalanceSheetMainGroup
            object as string. Else, empty string will be returned.
        end : str
            Determines the string to append at the end of the


        Example
        -------
        >> bsmg = BalanceSheetMainGroup(id=1, name="Asset")
        >>> bsmg.to_csv()
        '1,Asset'
        >>> bsmg.to_csv(-1)
        ''
        >>> bsmg.to_csv(end="\n")
        >>>
        '1,Asset
        '
        """
        assert type(depth) is int
        if depth < 0:
            return "" + end
        else:
            return f"{self.id},{self.name}" + end

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BalanceSheetMainGroup):
            return False
        return self.id == other.id


# Balance Sheet Main Group Objects
ASSET = BalanceSheetMainGroup(
    id=1,
    name="Asset",
)
LIABILITY = BalanceSheetMainGroup(
    id=2,
    name="Liability",
)
EQUITY = BalanceSheetMainGroup(
    id=3,
    name="Equity",
)

BALANCE_SHEET_MAIN_GROUP_OBJECTS = [
    ASSET,
    LIABILITY,
    EQUITY,
]


class BalanceSheetMainGroupObjects:
    """Class with class level attributes which contains
    standard balance sheet main group objects.

    Attributes
    ----------
    ASSET : BalanceSheetMainGroup
    LIABILITY : BalanceSheetMainGroup
    EQUITY : BalanceSheetMainGroup
    """

    ASSET = ASSET
    LIABILITY = LIABILITY
    EQUITY = EQUITY



# Import required functions
from pathlib import Path

from setuptools import find_packages, setup

import xarizmi

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


# Call setup function
setup(
    author="Javad Ebadi",
    author_email="javad@javadebadi.com",
    description=(
        "Xarizmi (read Khwarizmi) project is an educational project that"
        "contains tools for technical analysis in Python."
    ),
    name="xarizmi",
    include_package_data=True,  # Include files specified in MANIFEST.in
    package_data={
        "xarizmi.db": ["alembic.ini"],  # Include alembic.ini
        "xarizmi.db.alembic": ["*", "versions/*"],  # Include migrations
    },
    packages=find_packages(include=["xarizmi", "xarizmi.*"]),
    version=xarizmi.__version__,
    install_requires=[
        "pydantic==2.7.0",
        "scipy>=1.13.1",
        "numpy~=1.24",
        "mplfinance",
        "matplotlib",
        "yfinance~=0.2",
        "TA-Lib",
        "alembic>=1.13.1",
        "SQLAlchemy>=2.0.30",
    ],
    python_requires=">=3.11",
    license="Apache 2.0",
    url="https://github.com/javadebadi/xarizmi",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



"""A module to calculate time value of money
"""


def risk_free_return(
    impatience_to_consume: float,
    inflation: float,
) -> float:
    """Calculates risk free return of time value money.

    Parameters
    ----------
    impatience_to_consume : float
        People tend to prefer consuming now rather than later.
        This is called impatience to consume.
        The other name for this in finance is "pure rate of interest".
        For example, a person might have 2 percent (0.02) Impatience to
        consume which means that he or she expects 2 percent reward in order
        to not consume and invest instead.

    inflation : float
        The inflation as defined in economy which is general progressive
        increase in prices of goods and services in an economy.


    Returns
    -------
    float
        Returns the risk free return


    Example
    -------
    >>> risk_free_return(0.05,0.10)
    0.155

    """
    return (1 + impatience_to_consume) * (1 + inflation) - 1


def time_value_money(
    impatience_to_consume: float,
    inflation: float,
    risk: float,
) -> float:
    """Calculates time value of money.

    Parameters
    ----------
    impatience_to_consume : float
        People tend to prefer consuming now rather than later.
        This is called impatience to consume.
        The other name for this in finance is "pure rate of interest".
        For example, a person might have 2 percent (0.02) Impatience to
        consume which means that he or she expects 2 percent reward in order
        to not consume and invest instead.

    inflation : float
        The inflation as defined in economy which is general progressive
        increase in prices of goods and services in an economy.

    risk : float
        The amount of risk which exists in an investment.


    Returns
    -------
    float
        Returns the time value of money.


    Example
    -------
    >>> round(time_value_money(0.05,0.10,0.045),4)
    0.2
    >>> round(time_value_money(0.02,0.03,0),4)
    0.0506

    """
    return (1 + impatience_to_consume) * (1 + inflation) * (1 + risk) - 1


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



"""add_order_model

Revision ID: 40610635cc37
Revises: e33464855650
Create Date: 2025-04-08 18:17:09.117814

"""

from typing import Sequence
from typing import Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "40610635cc37"
down_revision: Union[str, None] = "e33464855650"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "xarizmi_order",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("symbol_id", sa.Integer(), nullable=False),
        sa.Column("order_id", sa.String(), nullable=False),
        sa.Column("amount", sa.Float(), nullable=False),
        sa.Column("price", sa.Float(), nullable=False),
        sa.Column(
            "side", sa.Enum("BUY", "SELL", name="side_enum"), nullable=False
        ),
        sa.Column(
            "status",
            sa.Enum(
                "ACTIVE",
                "DONE",
                "PARTIALLY_FILLED",
                "CANCELLED",
                name="order_status_enum",
            ),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["symbol_id"],
            ["xarizmi_symbol.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "symbol_id", "order_id", name="uix_symbol_id_order_id"
        ),
    )
    op.alter_column(
        "xarizmi_candlestick",
        "interval_type",
        existing_type=postgresql.INTERVAL(),
        type_=sa.Enum(
            "MIN_1",
            "MIN_3",
            "MIN_5",
            "MIN_15",
            "MIN_30",
            "HOUR_1",
            "HOUR_2",
            "HOUR_4",
            "HOUR_6",
            "HOUR_8",
            "HOUR_12",
            "DAY_1",
            "DAY_7",
            "DAY_14",
            "WEEK_1",
            "MONTH_1",
            name="interval_type_enum",
        ),
        existing_nullable=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "xarizmi_candlestick",
        "interval_type",
        existing_type=sa.Enum(
            "MIN_1",
            "MIN_3",
            "MIN_5",
            "MIN_15",
            "MIN_30",
            "HOUR_1",
            "HOUR_2",
            "HOUR_4",
            "HOUR_6",
            "HOUR_8",
            "HOUR_12",
            "DAY_1",
            "DAY_7",
            "DAY_14",
            "WEEK_1",
            "MONTH_1",
            name="interval_type_enum",
        ),
        type_=postgresql.INTERVAL(),
        existing_nullable=False,
    )
    op.drop_table("xarizmi_order")
    # ### end Alembic commands ###



from .delete import delete_all_cancelled_orders
from .delete import delete_unique_order
from .read import get_active_orders
from .read import get_orders
from .read import get_unique_order
from .upsert import upsert_order

__all__ = [
    "get_unique_order",
    "get_orders",
    "upsert_order",
    "delete_unique_order",
    "delete_all_cancelled_orders",
    "get_active_orders",
]



from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure


class TimeSeriesLinePlot:

    def __init__(
        self,
        fig_size: tuple[int, int] = (10, 6),
        save_path: str | None = None,
        label: str = "",
        xlabel: str = "",
        ylabel: str = "",
        title: str = "",
        color: str = "blue",
    ) -> None:
        self.fig_size = fig_size
        self.save_path = save_path
        self.label = label
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.title = title
        self.color = color
        self.fig, self.ax = plt.subplots(1, 1, figsize=self.fig_size)

    def plot_main_data(
        self, data: list[float], dates_data: list[datetime | None]
    ) -> tuple[Figure, Axes]:
        if any(value is None for value in dates_data):
            self.ax.plot(data, label=self.label, color=self.color)
        else:
            self.ax.plot(
                dates_data,  # type: ignore
                data,
                label=self.label,
                color=self.color,
            )

        self.ax.set_title(self.title)
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        self.ax.legend()
        plt.grid(True)
        return self.fig, self.ax

    def highlight_points(
        self,
        data: list[float],
        dates_data: list[pd.Timestamp | datetime | None],
        color: str = "r",
        marker: str = "o",
        s: float = 100,
        label: str | None = None,
    ) -> tuple[Figure, Axes]:
        self.ax.scatter(
            dates_data,  # type: ignore
            data,
            color=color,
            marker=marker,
            s=s,
            label=label,
        )
        return self.fig, self.ax

    def save(self, save_path: str | None = None) -> None:
        if save_path:
            self.fig.savefig(save_path)
        elif self.save_path:
            self.fig.savefig(self.save_path)

    def show(self) -> None:
        plt.show()



from pydantic import BaseModel


class Currency(BaseModel):
    name: str

    def to_string(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Currency):
            raise NotImplementedError(
                f"Can't compare objects of type {type(other)} and {type(self)}"
            )
        return self.name == other.name



from enum import IntEnum
from enum import StrEnum


class SideEnum(StrEnum):
    BUY = "BUY"
    SELL = "SELL"


class OrderStatusEnum(StrEnum):
    ACTIVE = "ACTIVE"
    DONE = "DONE"
    PARTIALLY_FILLED = "PARTIALLY_FILLED"
    CANCELLED = "CANCELLED"


class TimeUnitsSeconds(IntEnum):
    MIN = 60
    HOUR = 60 * 60
    DAY = 60 * 60 * 24
    WEEK = 60 * 60 * 24 * 7
    MONTH = 60 * 60 * 24 * 7 * 30


class IntervalTypeEnum(StrEnum):
    MIN_1 = "1min"
    MIN_3 = "3min"
    MIN_5 = "5min"
    MIN_15 = "15min"
    MIN_30 = "30min"
    HOUR_1 = "1hour"
    HOUR_2 = "2hour"
    HOUR_4 = "4hour"
    HOUR_6 = "6hour"
    HOUR_8 = "8hour"
    HOUR_12 = "12hour"
    DAY_1 = "1day"
    DAY_7 = "7day"
    DAY_14 = "14day"
    WEEK_1 = "1week"
    MONTH_1 = "1month"

    @staticmethod
    def get_interval_in_seconds(
        interval_type: "IntervalTypeEnum",
    ) -> int:
        unit: str | None = None
        value: int | None = None
        for prefix in [
            "MIN_",
            "HOUR_",
            "DAY_",
            "WEEK_",
            "MONTH_",
        ]:
            if interval_type.name.startswith(prefix):
                unit = prefix.replace("_", "")
                value = int(interval_type.name.split(prefix)[-1])
                break
        if unit is not None and value is not None:
            return value * TimeUnitsSeconds[unit].value
        else:
            raise NotImplementedError(
                f"Not implemented for interval_type = '{interval_type}'"
            )

    @staticmethod
    def get_interval_in_miliseconds(
        interval_type: "IntervalTypeEnum",
    ) -> int:
        return 1000 * IntervalTypeEnum.get_interval_in_seconds(interval_type)

    @staticmethod
    def get_interval_in_nanoseconds(
        interval_type: "IntervalTypeEnum",
    ) -> int:
        return 1000000000 * IntervalTypeEnum.get_interval_in_seconds(
            interval_type
        )



"""Module for sub main group of items in Balance Sheet
"""

from pydantic import BaseModel

from .balance_sheet_main_group import BalanceSheetMainGroup
from .balance_sheet_main_group import BalanceSheetMainGroupObjects as BSMain


class BalanceSheetSubMainGroup(BaseModel):
    """
    Class to determine Balance Sheet Sub Group.

    Balance Sheet Sub Main Group is one level more granular classification
    of items in balances sheet.

    Balance Sheet Main Group and Sub Main groups are:
    - Asset
        - Current asset
        - Long-term investment
        - Property, plant, and equipment
        - Intangible asset
    - Liability
        - Current Liability
        - Long-term liability
    - Equity
        - Owner's Investment

    """

    id: int
    name: str
    balance_sheet_main_group: BalanceSheetMainGroup

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"BalanceSheetMainGroup(id={self.id}, name={self.name})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BalanceSheetSubMainGroup):
            return False
        return self.id == other.id

    def to_csv(self, depth: int = 0, end: str = "") -> str:
        """Returns comma separated value representation."""
        if depth < 0:
            return "" + end
        deep_part = f"{self.balance_sheet_main_group.to_csv(depth=depth-1)}"
        if deep_part:
            deep_part = "," + deep_part
        return f"{self.id},{self.name}" + deep_part + end


# Balance Sheet Sub Main Group Objects

SHORT_TERM_ASSET = BalanceSheetSubMainGroup(
    id=1,
    name="Short Term Asset",
    balance_sheet_main_group=BSMain.ASSET,
)

LONG_TERM_ASSET = BalanceSheetSubMainGroup(
    id=2,
    name="Long Term Asset",
    balance_sheet_main_group=BSMain.ASSET,
)

SHORT_TERM_LIABILITY = BalanceSheetSubMainGroup(
    id=11,
    name="Short Term Liability",
    balance_sheet_main_group=BSMain.LIABILITY,
)

LONG_TERM_LIABILITY = BalanceSheetSubMainGroup(
    id=12,
    name="Long Term Liability",
    balance_sheet_main_group=BSMain.LIABILITY,
)

OWNERS_EQUITY = BalanceSheetSubMainGroup(
    id=21,
    name="Owner's Equity",
    balance_sheet_main_group=BSMain.EQUITY,
)


class BalanceSheetSubMainGroupObjects:
    """Class with class level attributes which contains
    standard balance sheet submain group objects.

    Attributes
    ----------
    SHORT_TERM_ASSET : BalanceSheetSubMainGroup
    LONG_TERM_ASSET : BalanceSheetSubMainGroup
    SHORT_TERM_LIABILITY : BalanceSheetSubMainGroup
    LONG_TERM_LIABILITY : BalanceSheetSubMainGroup
    OWNERS_EQUITY : BalanceSheetSubMainGroup
    """

    SHORT_TERM_ASSET = SHORT_TERM_ASSET
    LONG_TERM_ASSET = LONG_TERM_ASSET
    SHORT_TERM_LIABILITY = SHORT_TERM_LIABILITY
    LONG_TERM_LIABILITY = LONG_TERM_LIABILITY
    OWNERS_EQUITY = OWNERS_EQUITY



from contextlib import contextmanager

from sqlalchemy import Engine
from sqlalchemy.exc import SQLAlchemyError

from xarizmi.config import get_config


@contextmanager
def session_scope():  # type: ignore
    """Provide a transactional scope around a series of operations."""
    session = get_config().session_maker()
    try:
        yield session
        session.commit()
    except SQLAlchemyError:
        session.rollback()
        raise
    finally:
        session.close()


def get_engine() -> Engine:
    return get_config().db_engine
